#ifndef POISSON_H
#define POISSON_H

#ifdef DOUBLE
#define doubleTex int2
#define doubleTex2 int4
#define Float double
#else
#define doubleTex double
#define doubleTex2 double2
#define Float float
#endif

using namespace std;

typedef struct
{
	double memoryAllocation;
	double memoryRelease;
	double textureBind;
	double textureUnbind;
	double host2device;
	double device2host;
	double device2device;
	double prepareData;
	double kernelLaplace;
	double kernelCharges;
	double kernelBoundaries;
	double kernelDielectricBoundaries;
	double kernelSaveDielectricBoundaries;
	double kernelCompareDielectricBoundaries;
	double others;
    double updatePointers;
}CUDA_profile;

/*
 * CUDA FUNCTIONS
 */
void cu_getDeviceInfo();
void cu_init(int device);
void cu_finish();

void cu_poisson_initPhi( const int nx, const int ny, const int nz, CUDA_profile &profile);
    
void cu_poisson_initComparePhi(const int sizeEven, const int sizeOdd, CUDA_profile &profile);

void cu_poisson_copyPhiToGPU(const int nx, const int ny, const int nz, double *phiEven, double *phiOdd, CUDA_profile &profile);

void cu_poisson_copyPhiFromGPU(const int nx, const int ny, const int nz, double *phiEven, double *phiOdd, CUDA_profile &profile);

void cu_poisson_initCopyCharges(const int nx, const int ny, const int nz, const int linear, const int nChargesEven, 
	int *chargesIndexEven, double *chargesEven, const int nChargesOdd, int *chargesIndexOdd, double *chargesOdd, CUDA_profile &profile);

void cu_poisson_initCopyDielectricBoundaries(const int nx, const int ny, const int nz,
	const int nBoundariesEven, int *boundariesPositionEven, double *boundariesCorrectionEven,
	const int nBoundariesOdd, int *boundariesPositionOdd, double *boundariesCorrectionOdd,
	const int linear, CUDA_profile &profile);

void cu_poisson_initCopyDivisor(const int sizeEven, const int sizeOdd, double *even_divisor, double *odd_divisor, CUDA_profile &profile);

void cu_poisson_destroyMemory(int usingDivisor, int usingNonLinear, CUDA_profile &profile);

void cu_poisson_runLaplace(const int nx, const int ny, const int nz, int color, const double c1, const double c2, CUDA_profile &profile);

void cu_poisson_runBoltzmann(const int nx, const int ny, const int nz, int linear, int color, const double c2, CUDA_profile &profile);

void cu_poisson_runCharges(const int nCharges, int color,CUDA_profile &profile);

void cu_poisson_runDielectricBoundaries(const int nx, const int ny, const int nz,const int nBoundaries, int color,CUDA_profile &profile);

void cu_poisson_runSavePhimap(const int ntp, int color, CUDA_profile &profile);

void cu_poisson_runSaveLinearConvergenceCheck(const int nx, const int ny, const int nz,const int nBoundaries, int color,CUDA_profile &profile);

double cu_poisson_runComparePhi(int ntp, int color, CUDA_profile &profile);

double cu_poisson_runLinearCheckConvergence(const int nx, const int ny, const int nz,const int nBoundaries, int color,CUDA_profile &profile);

void cu_poisson_initQmap(const int nx, const int ny, const int nz,int *debmap1, int *debmap2,double *qmap1, double *qmap2, CUDA_profile &profile);

void cu_poisson_updatePointers(const int nx, const int ny, const int nz, double omcomp,int numberEvenCharges, int numberOddCharges,
        int nEvenDielectricBoundaries, int nOddDielectricBoundaries,CUDA_profile &profile);

void cu_poisson_updateCharges(const int nCharges, int color, double omcomp,CUDA_profile &profile);

void cu_poisson_updateQmap(int nt, int parity,double fac1, double chi2, double chi3, double chi4, double chi5,CUDA_profile &profile);

void cu_poisson_copySegmentFromGpu(const int color, const int offset, const int nn, double *phi, CUDA_profile &profile);

void cu_poisson_copySegmentToGpu(const int color, const int offset, const int nn, double *phi, CUDA_profile &profile);

#endif /* CUDA */

