#ifndef CUDA_H_
#define CUDA_H_

/*
 * INCLUDES
 */

#include <cstdio>
#include <unistd.h>
#include <cuda.h>
#include <cuda_runtime_api.h>
#include <iostream>
#define cutilSafeCall(err)       __cuda_safeCall      (err, __FILE__, __LINE__)
///** Macro to check the last CUDA error */
#define cutilCheckMsg(msg) __cuda_checkLastError(msg, __FILE__, __LINE__)

/*
 * CUDA functions
 */
inline void __cuda_safeCall(cudaError err, const char *file, const int line)
{
    if (cudaSuccess != err)
    {
    	printf("System error - %s(%i) : cuda_safeCall() Runtime API error %d : %s.",file, line, (int)err, cudaGetErrorString(err));
        throw (int)err;
    }
}


/** Function to get the fastest gpu devise
 *  *  @param[out] cutGetMaxGflopsDeviceId	Fastest GPU device number.
 *   */
inline int cutGetMaxGflopsDeviceId()
{
	int device_count = 0;
	cudaGetDeviceCount( &device_count );

	cudaDeviceProp device_properties;
	int max_gflops_device = 0;
	int max_gflops = 0;
	
	int current_device = 0;
	cudaGetDeviceProperties( &device_properties, current_device );
	max_gflops = device_properties.multiProcessorCount * device_properties.clockRate;
	++current_device;

	while( current_device < device_count )
	{
		cudaGetDeviceProperties( &device_properties, current_device );
		int gflops = device_properties.multiProcessorCount * device_properties.clockRate;
		if( gflops > max_gflops )
		{
			max_gflops        = gflops;
			max_gflops_device = current_device;
		}
		++current_device;
	}

	return max_gflops_device;
}

void cuda_getDeviceInfo();
void cuda_init(int device);
void cuda_finish();

#endif /* CUDA_H_ */
