
#include <iostream>
#include <cudaPbsolver.h>
#include <cudaaux.h>
#include <cstdio>


using namespace std;


texture<doubleTex, 1, cudaReadModeElementType> cu_phiTex;
texture<doubleTex, 1, cudaReadModeElementType> cu_phiTex2;

double *cu_phiEven;
double *cu_phiOdd;
double *cu_phiEven2;
double *cu_phiOdd2;
double *cu_error;

	/* Charges */
texture<int, 1, cudaReadModeElementType> cu_chargesPositionTex;
texture<doubleTex, 1, cudaReadModeElementType> cu_chargesTex;

int *cu_chargesPositionEven;
double *cu_chargesEven;

int *cu_chargesPositionOdd;
double *cu_chargesOdd;

	/* Dielectric Boundaries */
texture<int, 1, cudaReadModeElementType> cu_boundariesPositionTex;
texture<doubleTex2, 1, cudaReadModeElementType> cu_boundariesFactor12Tex;
texture<doubleTex2, 1, cudaReadModeElementType> cu_boundariesFactor34Tex;
texture<doubleTex2, 1, cudaReadModeElementType> cu_boundariesFactor56Tex;

int *cu_boundariesPositionEven;
double *cu_boundariesFactor12Even;
double *cu_boundariesFactor34Even;
double *cu_boundariesFactor56Even;

int *cu_boundariesPositionOdd;
double *cu_boundariesFactor12Odd;
double *cu_boundariesFactor34Odd;
double *cu_boundariesFactor56Odd;

	/* Temporal variables for convergence test */
double *cu_convergenceEvenTmp;
double *cu_convergenceOddTmp;
double *cu_convergenceEvenError;
double *cu_convergenceOddError;

	/* Divisor */
// texture<unsigned char, 1, cudaReadModeElementType> cu_divisorTex;
// unsigned char *cu_divisor;
texture<doubleTex, 1, cudaReadModeElementType> cu_divisorTex;
double *cu_even_divisor;
double *cu_odd_divisor;

texture<doubleTex, 1, cudaReadModeElementType> cu_qmapTex;
double *cu_qmap1;
double *cu_qmap2;

texture<int, 1, cudaReadModeElementType> cu_debTex;
int *cu_debmap1;
int *cu_debmap2;


/*
 * AUX functions
 */

int iDivUp(int a, int b)
{
    return ((a % b) != 0) ? (a / b + 1) : (a / b);
}

__global__ void cu_poisson_kernelSavePhi(int nn, double *phi, int perThread)
{

    int p0 = (threadIdx.x + blockIdx.x * blockDim.x)*perThread;
	
    double phi2;
    int2 value;

	for (int pp = p0; pp < p0+perThread; pp++)
	{
		if (pp < nn)
		{
#ifdef DOUBLE
			value = tex1Dfetch(cu_phiTex, pp);
			phi2 = __hiloint2double(value.y, value.x);
#else
			phi2 = tex1Dfetch(cu_phiTex, pp);
#endif
			phi[pp] = phi2;
		}
	}

}

__global__ void cu_poisson_kernelComparePhi(const int ntp, double *cu_error, const int perThread)
{

	int p0 = (threadIdx.x + blockIdx.x * blockDim.x)*perThread;
	int tid = threadIdx.x;
	
	int2 value;
	double localMax;
	double localValue, localValue2;
	double localError;
	extern __shared__ int sharedMemory[];
	double *error = (double*)&sharedMemory[0];


	localMax = 0.0;

	for (int pp = p0; pp < p0+perThread; pp++)
	{
		if (pp < ntp)
		{
	#ifdef DOUBLE
			value = tex1Dfetch(cu_phiTex, pp);
			localValue = __hiloint2double(value.y, value.x);
			value = tex1Dfetch(cu_phiTex2, pp);
			localValue2 = __hiloint2double(value.y, value.x);
	#else
			localValue = tex1Dfetch(cu_phiTex, pp);
			localValue2 = tex1Dfetch(cu_phiTex2, pp);
	#endif


			localError = fabs(localValue - localValue2);
			if (localError > localMax)
				localMax = localError;
			
		}
		else
			break;	
	}

	error[tid] = localMax;
	__syncthreads();
	
	if (tid < 128) {error[tid] = (error[tid] > error[tid+128]) ? error[tid] : error[tid+128];}__syncthreads();
	if (tid < 64) {error[tid] = (error[tid] > error[tid+64]) ? error[tid] : error[tid+64];}__syncthreads();
	if (tid < 32) {error[tid] = (error[tid] > error[tid+32]) ? error[tid] : error[tid+32];}//__syncthreads();
	if (tid < 16) {error[tid] = (error[tid] > error[tid+16]) ? error[tid] : error[tid+16];}//__syncthreads();
	if (tid < 8) {error[tid] = (error[tid] > error[tid+8]) ? error[tid] : error[tid+8];}//__syncthreads();
	if (tid < 4) {error[tid] = (error[tid] > error[tid+4]) ? error[tid] : error[tid+4];}//__syncthreads();
	if (tid < 2) {error[tid] = (error[tid] > error[tid+2]) ? error[tid] : error[tid+2];}//__syncthreads();
	if (tid < 1) {error[tid] = (error[tid] > error[tid+1]) ? error[tid] : error[tid+1];}//__syncthreads();

	
	if (tid == 0 )
		cu_error[blockIdx.x] = error[0];
	
	return;
}

__global__ void cu_poisson_kernelUpdateDivisor(const int nt, double *divisor, double omcomp, int perThread)
{
    int pos;

    for (int ii=0; ii<perThread; ii++)
    {
    	pos = (threadIdx.x + blockIdx.x * blockDim.x)*perThread + ii;

	    if (pos >= nt)
		    return;

	    divisor[pos] *= omcomp;
    }
}


__global__ void cu_poisson_kernelUpdateQmap(const int nt, double fac1, 
	double chi2, double chi3, double chi4, double chi5, double *qmap, int perThread)
{
	int pos;

	for (int ii=0; ii<perThread; ii++)
	{
		pos = (threadIdx.x + blockIdx.x * blockDim.x)*perThread + ii;

	    if (pos >= nt)
		    return;

	    double f1,f3,tmp,phi;
	    int2 value;

	    int idebmap = tex1Dfetch(cu_debTex, pos);
	    
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, pos);
		phi = __hiloint2double(value.y, value.x);
#else
		phi = tex1Dfetch(cu_phiTex, pos);
#endif
			
	    if (idebmap==1)
	    {
	    	if (chi2*chi4<1.e-6)
			{
				f1 = phi;
				f3 = f1*f1;
				tmp = fac1*f3*f1*(chi3 + f3*chi5);
				if (f3 > 2500.0)
				    tmp = fac1*2500.0*50.0*(chi3+2500.0*chi5);
			}
			else
			{
				f1 = phi;
				f3 = f1*f1;
				tmp = fac1*f3*(chi2+f1*(chi3+f1*(chi4+f1*chi5)));
				if (f3 > 2500.0)
				    tmp = fac1*2500.0*(chi2+50.0*(chi3+50.0*(chi4+50.0*chi5)));
			}
	    }
	    else
	    	tmp = 0.0;

	    qmap[pos] = tmp;
	}

	return;
}


__global__ void cu_poisson_kernelUpdateCharges(double *qval, const int nCharges, double omcomp, int perThread)
{
	int pos;
	for (int ii=0; ii<perThread; ii++)
	{
		pos = (threadIdx.x + blockIdx.x * blockDim.x)*perThread + ii;
	    if (pos >= nCharges)
		    break;

	    qval[pos] *= omcomp;		
	}

	return;
}


__global__ void cu_poisson_kernelUpdateDielectrics(const int nBoundaries, double omcomp, 
	double *factor12, double *factor34, double *factor56, int perThread)
{
	int pos;
	for (int ii=0; ii<perThread; ii++)
	{
		pos = (threadIdx.x + blockIdx.x * blockDim.x)*perThread + ii;
	    if (pos >= 2*nBoundaries)
		    break;

	    factor12[pos] *= omcomp;
	    factor34[pos] *= omcomp;
	    factor56[pos] *= omcomp;		
	}

	return;

}


__global__ void cu_poisson_kernelLaplaceEven(const int nx, const int ny, const int nz, const int nxy,
	double *phiEven,	const double c1, const double c2)
{
	extern __shared__ int sharedMemory[];
	double *phiShared = (double*)&sharedMemory[0];

	int xs = threadIdx.x;
	int ys = threadIdx.y;
	int x = threadIdx.x + blockIdx.x * (blockDim.x - 2);
	int y = threadIdx.y + blockIdx.y * (blockDim.y - 2);

	if (x >= nx || y >= ny)
		return;

	int interior;
	int copyTop = (x + y) & 0x1;
	int z;
	int zInitEven = 2 - ((x + y) & 0x1);
	int zInitOdd = 1 + ((x + y) & 0x1);
	int nsx = blockDim.x;
	int nsy = blockDim.y;
	int indexEven;
	int indexOdd;
	int indexShared;
	int offset_zn = -nxy / 2 - 1;
	int offset_zp = nxy / 2;
	double top;
	double bottom;

#ifdef DOUBLE
	int2 value;
#endif

	if (x > 0 && y > 0 && x < nx - 1 && y < ny - 1 &&
		xs > 0 && ys > 0 && xs < nsx - 1 && ys < nsy - 1)
		interior = 1;
	else
		interior = 0;

	indexEven = (x + y * nx + zInitEven * nxy) / 2;
	indexOdd = (x + y * nx + zInitOdd * nxy) / 2;
	indexShared = xs + ys * nsx;

	if (interior)
	{
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, indexEven + offset_zn);
		bottom = __hiloint2double(value.y, value.x);
#else
		bottom = tex1Dfetch(cu_phiTex, indexEven + offset_zn);
#endif
	}

	for (z = zInitEven; z < nz; z += 2)
	{
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, indexOdd);
		phiShared[indexShared] = __hiloint2double(value.y, value.x);
#else
		phiShared[indexShared] = tex1Dfetch(cu_phiTex, indexOdd);
#endif
		__syncthreads();

		if (interior && z < nz - 1)
		{
			if (copyTop)
			{
				top = phiShared[indexShared];
			}
			else
			{
#ifdef DOUBLE
				value = tex1Dfetch(cu_phiTex, indexEven + offset_zp);
				top = __hiloint2double(value.y, value.x);
#else
				top = tex1Dfetch(cu_phiTex, indexEven + offset_zp);
#endif
			}

			phiEven[indexEven] =
				c2 * phiEven[indexEven] + c1 * (
					phiShared[indexShared - 1] +
					phiShared[indexShared + 1] +
					phiShared[indexShared - nsx] +
					phiShared[indexShared + nsx] +
					bottom + top);

			bottom = top;
		}

		__syncthreads();

		indexEven += nxy;
		indexOdd += nxy;
	}
}

__global__ void cu_poisson_kernelLaplaceOdd(
	const int nx, const int ny, const int nz, const int nxy,
	double *phiOdd,
	const double c1, const double c2)
{
	extern __shared__ int sharedMemory[];
	double *phiShared = (double*)&sharedMemory[0];

	int xs = threadIdx.x;
	int ys = threadIdx.y;
	int x = threadIdx.x + blockIdx.x * (blockDim.x - 2);
	int y = threadIdx.y + blockIdx.y * (blockDim.y - 2);

	if (x >= nx || y >= ny)
		return;

	int interior;
	int copyTop = 1 - (x + y) & 0x1;
	int z;
	int zInitEven = 2 - ((x + y) & 0x1);
	int zInitOdd = 1 + ((x + y) & 0x1);
	int nsx = blockDim.x;
	int nsy = blockDim.y;
	int indexEven;
	int indexOdd;
	int indexShared;
	int offset_zn = -nxy / 2;
	int offset_zp = nxy / 2 + 1;
	double top;
	double bottom;
#ifdef DOUBLE
	int2 value;
#endif

	if (x > 0 && y > 0 && x < nx - 1 && y < ny - 1 &&
		xs > 0 && ys > 0 && xs < nsx - 1 && ys < nsy - 1)
		interior = 1;
	else
		interior = 0;

	indexEven = (x + y * nx + zInitEven * nxy) / 2;
	indexOdd = (x + y * nx + zInitOdd * nxy) / 2;
	indexShared = xs + ys * nsx;

	if (interior)
	{
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, indexOdd + offset_zn);
		bottom = __hiloint2double(value.y, value.x);
#else
		bottom = tex1Dfetch(cu_phiTex, indexOdd + offset_zn);
#endif
	}

	for (z = zInitOdd; z < nz; z += 2)
	{
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, indexEven);
		phiShared[indexShared] = __hiloint2double(value.y, value.x);
#else
		phiShared[indexShared] = tex1Dfetch(cu_phiTex, indexEven);
#endif
		__syncthreads();

		if (interior && z < nz - 1)
		{
			if (copyTop)
			{
				top = phiShared[indexShared];
			}
			else
			{
#ifdef DOUBLE
				value = tex1Dfetch(cu_phiTex, indexOdd + offset_zp);
				top = __hiloint2double(value.y, value.x);
#else
				top = tex1Dfetch(cu_phiTex, indexOdd + offset_zp);
#endif
			}

			phiOdd[indexOdd] =
				c2 * phiOdd[indexOdd] + c1 * (
					phiShared[indexShared - 1] +
					phiShared[indexShared + 1] +
					phiShared[indexShared - nsx] +
					phiShared[indexShared + nsx] +
					bottom + top);

			bottom = top;
		}

		__syncthreads();

		indexEven += nxy;
		indexOdd += nxy;
	}
}

__global__ void cu_poisson_kernelBoltzmannEven_nl( const int nx, const int ny, const int nz, const int nxy,
	double *phiEven, const double c2 )
{
	extern __shared__ int sharedMemory[];
	double *phiShared = (double*)&sharedMemory[0];

	int xs = threadIdx.x;
	int ys = threadIdx.y;
	int x = threadIdx.x + blockIdx.x * (blockDim.x - 2);
	int y = threadIdx.y + blockIdx.y * (blockDim.y - 2);

	if (x >= nx || y >= ny)
		return;

	int interior;
	int copyTop = (x + y) & 0x1;
	int z;
	int zInitEven = 2 - ((x + y) & 0x1);
	int zInitOdd = 1 + ((x + y) & 0x1);
	int nsx = blockDim.x;
	int nsy = blockDim.y;
	int indexEven;
	int indexOdd;
	int indexShared;
	int offset_zn = -nxy / 2 - 1;
	int offset_zp = nxy / 2;
	double top;
	double bottom;
	double qmap;

#ifdef DOUBLE
	int2 value;
#endif

// 	unsigned char divisorFlag;
	double divisor;

	if (x > 0 && y > 0 && x < nx - 1 && y < ny - 1 && xs > 0 && ys > 0 && xs < nsx - 1 && ys < nsy - 1)
		interior = 1;
	else
		interior = 0;

	indexEven = (x + y * nx + zInitEven * nxy) / 2;
	indexOdd = (x + y * nx + zInitOdd * nxy) / 2;
	indexShared = xs + ys * nsx;

	if (interior)
	{
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, indexEven + offset_zn);
		bottom = __hiloint2double(value.y, value.x);
#else
		bottom = tex1Dfetch(cu_phiTex, indexEven + offset_zn);
#endif
	}

	for (z = zInitEven; z < nz; z += 2)
	{
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, indexOdd);
		phiShared[indexShared] = __hiloint2double(value.y, value.x);
#else
		phiShared[indexShared] = tex1Dfetch(cu_phiTex, indexOdd);
#endif
		__syncthreads();

		if (interior && z < nz - 1)
		{
			if (copyTop)
			{
				top = phiShared[indexShared];
			}
			else
			{
#ifdef DOUBLE
				value = tex1Dfetch(cu_phiTex, indexEven + offset_zp);
				top = __hiloint2double(value.y, value.x);
#else
				top = tex1Dfetch(cu_phiTex, indexEven + offset_zp);
#endif
			}

#ifdef DOUBLE
			value = tex1Dfetch(cu_divisorTex, indexEven);
			divisor = __hiloint2double(value.y, value.x);
#else
			divisor = tex1Dfetch(cu_divisorTex, indexEven);
#endif
			
#ifdef DOUBLE
			value = tex1Dfetch(cu_qmapTex, indexEven);
			qmap = __hiloint2double(value.y, value.x);
#else
			qmap = tex1Dfetch(cu_qmapTex, indexEven);
#endif			
			
			phiEven[indexEven] =
				c2 * phiEven[indexEven] + divisor * (qmap + 
					phiShared[indexShared - 1] +
					phiShared[indexShared + 1] +
					phiShared[indexShared - nsx] +
					phiShared[indexShared + nsx] +
					bottom + top);

			bottom = top;
		}

		__syncthreads();

		indexEven += nxy;
		indexOdd += nxy;
	}
}

 // Kernel por Non-linear Boltzmann 
__global__ void cu_poisson_kernelBoltzmannOdd_nl(
	const int nx, const int ny, const int nz, const int nxy,
	double *phiOdd, const double c2)
{
	extern __shared__ int sharedMemory[];
	double *phiShared = (double*)&sharedMemory[0];

	int xs = threadIdx.x;
	int ys = threadIdx.y;
	int x = threadIdx.x + blockIdx.x * (blockDim.x - 2);
	int y = threadIdx.y + blockIdx.y * (blockDim.y - 2);

	if (x >= nx || y >= ny)
		return;

	int interior;
	int copyTop = 1 - (x + y) & 0x1;
	int z;
	int zInitEven = 2 - ((x + y) & 0x1);
	int zInitOdd = 1 + ((x + y) & 0x1);
	int nsx = blockDim.x;
	int nsy = blockDim.y;
	int indexEven;
	int indexOdd;
	int indexShared;
	int offset_zn = -nxy / 2;
	int offset_zp = nxy / 2 + 1;
	double top;
	double bottom;
	double qmap;
#ifdef DOUBLE
	int2 value;
#endif
// 	unsigned char divisorFlag;
	double divisor;

	if (x > 0 && y > 0 && x < nx - 1 && y < ny - 1 &&
		xs > 0 && ys > 0 && xs < nsx - 1 && ys < nsy - 1)
		interior = 1;
	else
		interior = 0;

	indexEven = (x + y * nx + zInitEven * nxy) / 2;
	indexOdd = (x + y * nx + zInitOdd * nxy) / 2;
	indexShared = xs + ys * nsx;

	if (interior)
	{
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, indexOdd + offset_zn);
		bottom = __hiloint2double(value.y, value.x);
#else
		bottom = tex1Dfetch(cu_phiTex, indexOdd + offset_zn);
#endif
	}

	for (z = zInitOdd; z < nz; z += 2)
	{
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, indexEven);
		phiShared[indexShared] = __hiloint2double(value.y, value.x);
#else
		phiShared[indexShared] = tex1Dfetch(cu_phiTex, indexEven);
#endif
		__syncthreads();

		if (interior && z < nz - 1)
		{
			if (copyTop)
			{
				top = phiShared[indexShared];
			}
			else
			{
#ifdef DOUBLE
				value = tex1Dfetch(cu_phiTex, indexOdd + offset_zp);
				top = __hiloint2double(value.y, value.x);
#else
				top = tex1Dfetch(cu_phiTex, indexOdd + offset_zp);
#endif
			}	

#ifdef DOUBLE
			value = tex1Dfetch(cu_divisorTex, indexOdd);
			divisor = __hiloint2double(value.y, value.x);
#else
			divisor = tex1Dfetch(cu_divisorTex, indexOdd);
#endif

#ifdef DOUBLE
			value = tex1Dfetch(cu_qmapTex, indexOdd);
			qmap = __hiloint2double(value.y, value.x);
#else
			qmap = tex1Dfetch(cu_divisorTex, indexOdd);
#endif	
			phiOdd[indexOdd] =
				c2 * phiOdd[indexOdd] + divisor * ( qmap + 
					phiShared[indexShared - 1] +
					phiShared[indexShared + 1] +
					phiShared[indexShared - nsx] +
					phiShared[indexShared + nsx] +
					bottom + top);

			bottom = top;
		}

		__syncthreads();

		indexEven += nxy;
		indexOdd += nxy;
	}
}


/*
 * CUDA kernel Boltzmann
 */

__global__ void cu_poisson_kernelBoltzmannEven(const int nx, const int ny, const int nz, const int nxy, double *phiEven, const double c2)
{
	extern __shared__ int sharedMemory[];
	double *phiShared = (double*)&sharedMemory[0];

	int xs = threadIdx.x;
	int ys = threadIdx.y;
	int x = threadIdx.x + blockIdx.x * (blockDim.x - 2);
	int y = threadIdx.y + blockIdx.y * (blockDim.y - 2);

	if (x >= nx || y >= ny)
		return;

	int interior;
	int copyTop = (x + y) & 0x1;
	int z;
	int zInitEven = 2 - ((x + y) & 0x1);
	int zInitOdd = 1 + ((x + y) & 0x1);
	int nsx = blockDim.x;
	int nsy = blockDim.y;
	int indexEven;
	int indexOdd;
	int indexShared;
	int offset_zn = -nxy / 2 - 1;
	int offset_zp = nxy / 2;
	double top;
	double bottom;
#ifdef DOUBLE
	int2 value;
#endif
// 	unsigned char divisorFlag;
	double divisor;

	if (x > 0 && y > 0 && x < nx - 1 && y < ny - 1 &&
		xs > 0 && ys > 0 && xs < nsx - 1 && ys < nsy - 1)
		interior = 1;
	else
		interior = 0;

	indexEven = (x + y * nx + zInitEven * nxy) / 2;
	indexOdd = (x + y * nx + zInitOdd * nxy) / 2;
	indexShared = xs + ys * nsx;

	if (interior)
	{
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, indexEven + offset_zn);
		bottom = __hiloint2double(value.y, value.x);
#else
		bottom = tex1Dfetch(cu_phiTex, indexEven + offset_zn);
#endif
	}

	for (z = zInitEven; z < nz; z += 2)
	{
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, indexOdd);
		phiShared[indexShared] = __hiloint2double(value.y, value.x);
#else
		phiShared[indexShared] = tex1Dfetch(cu_phiTex, indexOdd);
#endif
		__syncthreads();

		if (interior && z < nz - 1)
		{
			if (copyTop)
			{
				top = phiShared[indexShared];
			}
			else
			{
#ifdef DOUBLE
				value = tex1Dfetch(cu_phiTex, indexEven + offset_zp);
				top = __hiloint2double(value.y, value.x);
#else
				top = tex1Dfetch(cu_phiTex, indexEven + offset_zp);
#endif
			}

#ifdef DOUBLE
			value = tex1Dfetch(cu_divisorTex, indexEven);
			divisor = __hiloint2double(value.y, value.x);
#else
			divisor = tex1Dfetch(cu_divisorTex, indexEven);
#endif
			
			phiEven[indexEven] =
				c2 * phiEven[indexEven] + divisor * 
					(phiShared[indexShared - 1] +
					phiShared[indexShared + 1] +
					phiShared[indexShared - nsx] +
					phiShared[indexShared + nsx] +
					bottom + top);

			bottom = top;
		}

		__syncthreads();

		indexEven += nxy;
		indexOdd += nxy;
	}
}

__global__ void cu_poisson_kernelBoltzmannOdd(
	const int nx, const int ny, const int nz, const int nxy,
	double *phiOdd, const double c2)
{
	extern __shared__ int sharedMemory[];
	double *phiShared = (double*)&sharedMemory[0];

	int xs = threadIdx.x;
	int ys = threadIdx.y;
	int x = threadIdx.x + blockIdx.x * (blockDim.x - 2);
	int y = threadIdx.y + blockIdx.y * (blockDim.y - 2);

	if (x >= nx || y >= ny)
		return;

	int interior;
	int copyTop = 1 - (x + y) & 0x1;
	int z;
	int zInitEven = 2 - ((x + y) & 0x1);
	int zInitOdd = 1 + ((x + y) & 0x1);
	int nsx = blockDim.x;
	int nsy = blockDim.y;
	int indexEven;
	int indexOdd;
	int indexShared;
	int offset_zn = -nxy / 2;
	int offset_zp = nxy / 2 + 1;
	double top;
	double bottom;
#ifdef DOUBLE
	int2 value;
#endif
// 	unsigned char divisorFlag;
	double divisor;

	if (x > 0 && y > 0 && x < nx - 1 && y < ny - 1 &&
		xs > 0 && ys > 0 && xs < nsx - 1 && ys < nsy - 1)
		interior = 1;
	else
		interior = 0;

	indexEven = (x + y * nx + zInitEven * nxy) / 2;
	indexOdd = (x + y * nx + zInitOdd * nxy) / 2;
	indexShared = xs + ys * nsx;

	if (interior)
	{
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, indexOdd + offset_zn);
		bottom = __hiloint2double(value.y, value.x);
#else
		bottom = tex1Dfetch(cu_phiTex, indexOdd + offset_zn);
#endif
	}

	for (z = zInitOdd; z < nz; z += 2)
	{
#ifdef DOUBLE
		value = tex1Dfetch(cu_phiTex, indexEven);
		phiShared[indexShared] = __hiloint2double(value.y, value.x);
#else
		phiShared[indexShared] = tex1Dfetch(cu_phiTex, indexEven);
#endif
		__syncthreads();

		if (interior && z < nz - 1)
		{
			if (copyTop)
			{
				top = phiShared[indexShared];
			}
			else
			{
#ifdef DOUBLE
				value = tex1Dfetch(cu_phiTex, indexOdd + offset_zp);
				top = __hiloint2double(value.y, value.x);
#else
				top = tex1Dfetch(cu_phiTex, indexOdd + offset_zp);
#endif
			}

#ifdef DOUBLE
			value = tex1Dfetch(cu_divisorTex, indexOdd);
			divisor = __hiloint2double(value.y, value.x);
#else
			divisor = tex1Dfetch(cu_divisorTex, indexOdd);
#endif
			
			phiOdd[indexOdd] =
				c2 * phiOdd[indexOdd] + divisor * (
					phiShared[indexShared - 1] +
					phiShared[indexShared + 1] +
					phiShared[indexShared - nsx] +
					phiShared[indexShared + nsx] +
					bottom + top);

			bottom = top;
		}

		__syncthreads();

		indexEven += nxy;
		indexOdd += nxy;
	}
}

/*
 * CUDA kernel Charges
 */

__global__ void cu_poisson_kernelCharges(double *phi, const int nCharges)
{
	int pos = threadIdx.x + blockIdx.x * blockDim.x;

	if (pos >= nCharges)
		return;

	int index = tex1Dfetch(cu_chargesPositionTex, pos);
	double newPhi = phi[index];

#ifdef DOUBLE
	int2 value = tex1Dfetch(cu_chargesTex, pos);
	newPhi += __hiloint2double(value.y, value.x);
#else
	newPhi += tex1Dfetch(cu_chargesTex, pos);
#endif

	phi[index] = newPhi;
}


/*
 * CUDA Kernel Dielectric Boundaries
 */

__global__ void cu_poisson_kernelDielectricBoundaries(
	const int nBoundaries, double *phi,
	const int offset_xn, const int offset_xp,
	const int offset_yn, const int offset_yp,
	const int offset_zn, const int offset_zp)
{
	int pos = threadIdx.x + blockIdx.x * blockDim.x;

	if (pos >= nBoundaries)
		return;

	int index = tex1Dfetch(cu_boundariesPositionTex, pos);
	doubleTex2 factors[3];
	double newPhi = phi[index];

#ifdef DOUBLE
	int2 values[6];
#endif

	factors[0] = tex1Dfetch(cu_boundariesFactor12Tex, pos);
	factors[1] = tex1Dfetch(cu_boundariesFactor34Tex, pos);
	factors[2] = tex1Dfetch(cu_boundariesFactor56Tex, pos);

#ifdef DOUBLE
	values[0] = tex1Dfetch(cu_phiTex, index + offset_xn);
	values[1] = tex1Dfetch(cu_phiTex, index + offset_xp);
	values[2] = tex1Dfetch(cu_phiTex, index + offset_yn);
	values[3] = tex1Dfetch(cu_phiTex, index + offset_yp);
	values[4] = tex1Dfetch(cu_phiTex, index + offset_zn);
	values[5] = tex1Dfetch(cu_phiTex, index + offset_zp);

	newPhi +=
		__hiloint2double(values[0].y, values[0].x) * __hiloint2double(factors[0].y, factors[0].x) +
		__hiloint2double(values[1].y, values[1].x) * __hiloint2double(factors[0].w, factors[0].z) +
		__hiloint2double(values[2].y, values[2].x) * __hiloint2double(factors[1].y, factors[1].x) +
		__hiloint2double(values[3].y, values[3].x) * __hiloint2double(factors[1].w, factors[1].z) +
		__hiloint2double(values[4].y, values[4].x) * __hiloint2double(factors[2].y, factors[2].x) +
		__hiloint2double(values[5].y, values[5].x) * __hiloint2double(factors[2].w, factors[2].z);
#else
	newPhi +=
		tex1Dfetch(cu_phiTex, index + offset_xn) * factors[0].x +
		tex1Dfetch(cu_phiTex, index + offset_xp) * factors[0].y +
		tex1Dfetch(cu_phiTex, index + offset_yn) * factors[1].x +
		tex1Dfetch(cu_phiTex, index + offset_yp) * factors[1].y +
		tex1Dfetch(cu_phiTex, index + offset_zn) * factors[2].x +
		tex1Dfetch(cu_phiTex, index + offset_zp) * factors[2].y;
#endif

	phi[index] = newPhi;
}

__global__ void cu_poisson_kernelSaveConvergenceCheck(
	const int nBoundaries, double *boundaries)
{
	int pos = threadIdx.x + blockIdx.x * blockDim.x;

	if (pos >= nBoundaries)
		return;

	int index = tex1Dfetch(cu_chargesPositionTex, pos);

#ifdef DOUBLE
	int2 value = tex1Dfetch(cu_phiTex, index);
	boundaries[pos] = __hiloint2double(value.y, value.x);
#else
	boundaries[pos] = tex1Dfetch(cu_phiTex, index);
#endif
}

__global__ void cu_poisson_kernelLinearConvergenceCheck(const int nBoundaries, double *boundaries, double *errors)
{
	int pos = threadIdx.x + blockIdx.x * blockDim.x;

	if (pos >= nBoundaries)
		return;

	int index = tex1Dfetch(cu_chargesPositionTex, pos);
	double boundary;

#ifdef DOUBLE
	int2 value = tex1Dfetch(cu_phiTex, index);
	boundary = __hiloint2double(value.y, value.x);
#else
	boundary = tex1Dfetch(cu_phiTex, index);
#endif

	errors[pos] = fabs(boundaries[pos] - boundary);
}


/*
 * CUDA functions
 */

void cu_poisson_initPhi(const int nx, const int ny, const int nz,CUDA_profile &profile)
{

	// Variables
	int sizeTotal;
	int sizeEven;
	int sizeOdd;
	struct timespec time0, time1;

	// Calculate size
	clock_gettime(CLOCK_REALTIME, &time0);
	sizeTotal = nx * ny * nz;
	sizeOdd = sizeTotal / 2;
	sizeEven = sizeTotal - sizeOdd;
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Memory allocation
	clock_gettime(CLOCK_REALTIME, &time0);
	cutilSafeCall(cudaMalloc((void**)&cu_phiEven, sizeEven * sizeof(double)));
	cutilSafeCall(cudaMalloc((void**)&cu_phiOdd, sizeOdd * sizeof(double)));
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.memoryAllocation += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Texture params init
	clock_gettime(CLOCK_REALTIME, &time0);

	cu_phiTex.normalized = false;
	cu_phiTex.filterMode = cudaFilterModePoint;
	cu_phiTex.addressMode[0] = cudaAddressModeClamp;

	cu_phiTex2.normalized = false;
	cu_phiTex2.filterMode = cudaFilterModePoint;
	cu_phiTex2.addressMode[0] = cudaAddressModeClamp;

	clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

}

void cu_poisson_copyPhiToGPU(
	const int nx, const int ny, const int nz,
	double *phiEven, double *phiOdd,
	CUDA_profile &profile)
{
	// Variables
	int sizeTotal;
	int sizeEven;
	int sizeOdd;

	struct timespec time0, time1;

	// Calculate size
	clock_gettime(CLOCK_REALTIME, &time0);
	sizeTotal = nx * ny * nz;
	sizeOdd = sizeTotal / 2;
	sizeEven = sizeTotal - sizeOdd;
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Copy data to graphic card
	clock_gettime(CLOCK_REALTIME, &time0);
	cutilSafeCall(cudaMemcpy(cu_phiEven, phiEven, sizeEven * sizeof(double), cudaMemcpyHostToDevice));
	cutilSafeCall(cudaMemcpy(cu_phiOdd, phiOdd, sizeOdd * sizeof(double), cudaMemcpyHostToDevice));
	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.host2device += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
}

void cu_poisson_copyPhiFromGPU(
	const int nx, const int ny, const int nz,
	double *phiEven, double *phiOdd,
	CUDA_profile &profile)
{
	// Variables
	int sizeTotal;
	int sizeEven;
	int sizeOdd;

	struct timespec time0, time1;

	// Calculate size
	clock_gettime(CLOCK_REALTIME, &time0);
	sizeTotal = nx * ny * nz;
	sizeOdd = sizeTotal / 2;
	sizeEven = sizeTotal - sizeOdd;
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Copy data to host
	clock_gettime(CLOCK_REALTIME, &time0);

	cutilSafeCall(cudaMemcpy(phiEven, cu_phiEven, sizeEven * sizeof(double), cudaMemcpyDeviceToHost));
	cutilSafeCall(cudaMemcpy(phiOdd, cu_phiOdd, sizeOdd * sizeof(double), cudaMemcpyDeviceToHost));

	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.device2host += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
}

void cu_poisson_copySegmentFromGpu(const int color, const int offset, const int nn, double *phi, CUDA_profile &profile)
{

	struct timespec time0, time1;

	// Copy data to host
	clock_gettime(CLOCK_REALTIME, &time0);

	if (color==0) {
		cutilSafeCall(cudaMemcpy(phi, cu_phiEven+offset, nn * sizeof(double), cudaMemcpyDeviceToHost));
	} else {
		cutilSafeCall(cudaMemcpy(phi, cu_phiOdd+offset, nn * sizeof(double), cudaMemcpyDeviceToHost)); }
	
	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.device2host += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
}

void cu_poisson_copySegmentToGpu(const int color, const int offset, const int nn, double *phi, CUDA_profile &profile)
{

	struct timespec time0, time1;

	// Copy data to host
	clock_gettime(CLOCK_REALTIME, &time0);

	if (color==0) {
		cutilSafeCall(cudaMemcpy(cu_phiEven+offset, phi, nn * sizeof(double), cudaMemcpyHostToDevice));
	} else {
		cutilSafeCall(cudaMemcpy(cu_phiOdd+offset, phi, nn * sizeof(double), cudaMemcpyHostToDevice)); }
	
	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.device2host += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
}

void cu_poisson_initCopyCharges(
	const int nx, const int ny, const int nz, const int linear,
	const int nChargesEven, int *chargesIndexEven, double *chargesEven,
	const int nChargesOdd, int *chargesIndexOdd, double *chargesOdd,
	CUDA_profile &profile)
{
	// Variables
	struct timespec time0, time1;

	// Memory allocation
	clock_gettime(CLOCK_REALTIME, &time0);

	cutilSafeCall(cudaMalloc((void**)&cu_chargesPositionEven, nChargesEven * sizeof(int)));
	cutilSafeCall(cudaMalloc((void**)&cu_chargesEven, nChargesEven * sizeof(double)));

	cutilSafeCall(cudaMalloc((void**)&cu_chargesPositionOdd, nChargesOdd * sizeof(int)));
	cutilSafeCall(cudaMalloc((void**)&cu_chargesOdd, nChargesOdd * sizeof(double)));

	clock_gettime(CLOCK_REALTIME, &time1);
	profile.memoryAllocation += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Copy data
	clock_gettime(CLOCK_REALTIME, &time0);

	cutilSafeCall(cudaMemcpy(cu_chargesPositionEven, chargesIndexEven, nChargesEven * sizeof(int), cudaMemcpyHostToDevice));
	cutilSafeCall(cudaMemcpy(cu_chargesEven, chargesEven, nChargesEven * sizeof(double), cudaMemcpyHostToDevice));

	cutilSafeCall(cudaMemcpy(cu_chargesPositionOdd, chargesIndexOdd, nChargesOdd * sizeof(int), cudaMemcpyHostToDevice));
	cutilSafeCall(cudaMemcpy(cu_chargesOdd, chargesOdd, nChargesOdd * sizeof(double), cudaMemcpyHostToDevice));

	clock_gettime(CLOCK_REALTIME, &time1);
	profile.host2device += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Texture params init
	clock_gettime(CLOCK_REALTIME, &time0);

	cu_chargesPositionTex.normalized = false;
	cu_chargesPositionTex.filterMode = cudaFilterModePoint;
	cu_chargesPositionTex.addressMode[0] = cudaAddressModeClamp;

	cu_chargesTex.normalized = false;
	cu_chargesTex.filterMode = cudaFilterModePoint;
	cu_chargesTex.addressMode[0] = cudaAddressModeClamp;

	clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
	clock_gettime(CLOCK_REALTIME, &time0);

	// Allocate memory for temporal dielectric boundaries (for convergence test on linear solver)
    if (linear)
    {
	    clock_gettime(CLOCK_REALTIME, &time0);
	    cutilSafeCall(cudaMalloc((void**)&cu_convergenceEvenTmp, nChargesEven * sizeof(double)));
	    cutilSafeCall(cudaMalloc((void**)&cu_convergenceOddTmp, nChargesOdd * sizeof(double)));
	    cutilSafeCall(cudaMalloc((void**)&cu_convergenceEvenError, nChargesEven * sizeof(double)));
	    cutilSafeCall(cudaMalloc((void**)&cu_convergenceOddError, nChargesOdd * sizeof(double)));
	    clock_gettime(CLOCK_REALTIME, &time1);
	    profile.memoryAllocation += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
	}

	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("Error @ cu_poisson_initCopyCharges: %s\n", cudaGetErrorString(err));

    clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

}

void cu_poisson_initCopyDielectricBoundaries(
	const int nx, const int ny, const int nz,
	const int nBoundariesEven, int *boundariesPositionEven, double *boundariesCorrectionEven,
	const int nBoundariesOdd, int *boundariesPositionOdd, double *boundariesCorrectionOdd,
	const int linear, CUDA_profile &profile)
{
	// Variables
    struct timespec time0, time1;

    double *boundariesFactor12Even_aux;
    double *boundariesFactor34Even_aux;
    double *boundariesFactor56Even_aux;

    double *boundariesFactor12Odd_aux;
    double *boundariesFactor34Odd_aux;
    double *boundariesFactor56Odd_aux;

    // Memory allocation
    clock_gettime(CLOCK_REALTIME, &time0);

    cutilSafeCall(cudaMalloc((void**)&cu_boundariesPositionEven, nBoundariesEven * sizeof(int)));
    cutilSafeCall(cudaMalloc((void**)&cu_boundariesFactor12Even, nBoundariesEven * sizeof(double) * 2));
    cutilSafeCall(cudaMalloc((void**)&cu_boundariesFactor34Even, nBoundariesEven * sizeof(double) * 2));
    cutilSafeCall(cudaMalloc((void**)&cu_boundariesFactor56Even, nBoundariesEven * sizeof(double) * 2));

    cutilSafeCall(cudaMalloc((void**)&cu_boundariesPositionOdd, nBoundariesOdd * sizeof(int)));
    cutilSafeCall(cudaMalloc((void**)&cu_boundariesFactor12Odd, nBoundariesOdd * sizeof(double) * 2));
    cutilSafeCall(cudaMalloc((void**)&cu_boundariesFactor34Odd, nBoundariesOdd * sizeof(double) * 2));
    cutilSafeCall(cudaMalloc((void**)&cu_boundariesFactor56Odd, nBoundariesOdd * sizeof(double) * 2));

    clock_gettime(CLOCK_REALTIME, &time1);
    profile.memoryAllocation += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Convert data
    clock_gettime(CLOCK_REALTIME, &time0);

    boundariesFactor12Even_aux = new double[nBoundariesEven * 2];
    boundariesFactor34Even_aux = new double[nBoundariesEven * 2];
    boundariesFactor56Even_aux = new double[nBoundariesEven * 2];

    boundariesFactor12Odd_aux = new double[nBoundariesOdd * 2];
    boundariesFactor34Odd_aux = new double[nBoundariesOdd * 2];
    boundariesFactor56Odd_aux = new double[nBoundariesOdd * 2];

    for (int i = 0; i < nBoundariesEven; i++)
    {
	    boundariesFactor12Even_aux[i * 2] = boundariesCorrectionEven[i*6];
	    boundariesFactor12Even_aux[i * 2 + 1] = boundariesCorrectionEven[1+i*6];

	    boundariesFactor34Even_aux[i * 2] = boundariesCorrectionEven[2+i*6];
	    boundariesFactor34Even_aux[i * 2 + 1] = boundariesCorrectionEven[3+i*6];

	    boundariesFactor56Even_aux[i * 2] = boundariesCorrectionEven[4+i*6];
	    boundariesFactor56Even_aux[i * 2 + 1] = boundariesCorrectionEven[5+i*6];
    }

    for (int i = 0; i < nBoundariesOdd; i++)
    {
	    boundariesFactor12Odd_aux[i * 2] = boundariesCorrectionOdd[i*6];
	    boundariesFactor12Odd_aux[i * 2 + 1] = boundariesCorrectionOdd[1+i*6];

	    boundariesFactor34Odd_aux[i * 2] = boundariesCorrectionOdd[2+i*6];
	    boundariesFactor34Odd_aux[i * 2 + 1] = boundariesCorrectionOdd[3+i*6];

	    boundariesFactor56Odd_aux[i * 2] = boundariesCorrectionOdd[4+i*6];
	    boundariesFactor56Odd_aux[i * 2 + 1] = boundariesCorrectionOdd[5+i*6];
    }

    clock_gettime(CLOCK_REALTIME, &time1);
    profile.prepareData += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Copy data
    clock_gettime(CLOCK_REALTIME, &time0);

    cutilSafeCall(cudaMemcpy(cu_boundariesPositionEven, boundariesPositionEven, nBoundariesEven * sizeof(int), cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(cu_boundariesFactor12Even, boundariesFactor12Even_aux, nBoundariesEven * sizeof(double) * 2, cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(cu_boundariesFactor34Even, boundariesFactor34Even_aux, nBoundariesEven * sizeof(double) * 2, cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(cu_boundariesFactor56Even, boundariesFactor56Even_aux, nBoundariesEven * sizeof(double) * 2, cudaMemcpyHostToDevice));

    cutilSafeCall(cudaMemcpy(cu_boundariesPositionOdd, boundariesPositionOdd, nBoundariesOdd * sizeof(int), cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(cu_boundariesFactor12Odd, boundariesFactor12Odd_aux, nBoundariesOdd * sizeof(double) * 2, cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(cu_boundariesFactor34Odd, boundariesFactor34Odd_aux, nBoundariesOdd * sizeof(double) * 2, cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(cu_boundariesFactor56Odd, boundariesFactor56Odd_aux, nBoundariesOdd * sizeof(double) * 2, cudaMemcpyHostToDevice));

    clock_gettime(CLOCK_REALTIME, &time1);
    profile.host2device += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Delete aux data
    clock_gettime(CLOCK_REALTIME, &time0);

    delete [] boundariesFactor12Even_aux;
    delete [] boundariesFactor34Even_aux;
    delete [] boundariesFactor56Even_aux;

    delete [] boundariesFactor12Odd_aux;
    delete [] boundariesFactor34Odd_aux;
    delete [] boundariesFactor56Odd_aux;

    clock_gettime(CLOCK_REALTIME, &time1);
    profile.prepareData += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Texture params init
    clock_gettime(CLOCK_REALTIME, &time0);

    cu_boundariesPositionTex.normalized = false;
    cu_boundariesPositionTex.filterMode = cudaFilterModePoint;
    cu_boundariesPositionTex.addressMode[0] = cudaAddressModeClamp;

    cu_boundariesFactor12Tex.normalized = false;
    cu_boundariesFactor12Tex.filterMode = cudaFilterModePoint;
    cu_boundariesFactor12Tex.addressMode[0] = cudaAddressModeClamp;

    cu_boundariesFactor34Tex.normalized = false;
    cu_boundariesFactor34Tex.filterMode = cudaFilterModePoint;
    cu_boundariesFactor34Tex.addressMode[0] = cudaAddressModeClamp;

    cu_boundariesFactor56Tex.normalized = false;
    cu_boundariesFactor56Tex.filterMode = cudaFilterModePoint;
    cu_boundariesFactor56Tex.addressMode[0] = cudaAddressModeClamp;

	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("Error @ cu_poisson_initCopyDielectricBoundaries: %s\n", cudaGetErrorString(err));

    clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
}

void cu_poisson_initCopyDivisor(const int sizeEven, const int sizeOdd, double *even_divisor, double *odd_divisor, CUDA_profile &profile)
{

    struct timespec time0, time1;

    // Memory allocation
    clock_gettime(CLOCK_REALTIME, &time0);
    cutilSafeCall(cudaMalloc((void**)&cu_even_divisor, sizeEven * sizeof(double)));
    cutilSafeCall(cudaMalloc((void**)&cu_odd_divisor, sizeOdd * sizeof(double)));
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.memoryAllocation += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Texture params init
    clock_gettime(CLOCK_REALTIME, &time0);
    cu_divisorTex.normalized = false;
    cu_divisorTex.filterMode = cudaFilterModePoint;
    cu_divisorTex.addressMode[0] = cudaAddressModeClamp;
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Copy data
    clock_gettime(CLOCK_REALTIME, &time0);
    cutilSafeCall(cudaMemcpy(cu_even_divisor, even_divisor, sizeEven * sizeof(double), cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(cu_odd_divisor, odd_divisor, sizeOdd * sizeof(double), cudaMemcpyHostToDevice));
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.host2device += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("Error @ cu_poisson_initCopyDivisor: %s\n", cudaGetErrorString(err));

}

void cu_poisson_destroyMemory(int usingDivisor, int linear, CUDA_profile &profile)
{
	// Variables
	struct timespec time0, time1;

	// Release memory
	clock_gettime(CLOCK_REALTIME, &time0);

	cutilSafeCall(cudaFree(cu_phiEven));

	cutilSafeCall(cudaFree(cu_phiOdd));

	cutilSafeCall(cudaFree(cu_chargesPositionEven));
	cutilSafeCall(cudaFree(cu_chargesEven));

	cutilSafeCall(cudaFree(cu_chargesPositionOdd));
	cutilSafeCall(cudaFree(cu_chargesOdd));

	cutilSafeCall(cudaFree(cu_boundariesPositionEven));
	cutilSafeCall(cudaFree(cu_boundariesFactor12Even));
	cutilSafeCall(cudaFree(cu_boundariesFactor34Even));
	cutilSafeCall(cudaFree(cu_boundariesFactor56Even));

	cutilSafeCall(cudaFree(cu_boundariesPositionOdd));
	cutilSafeCall(cudaFree(cu_boundariesFactor12Odd));
	cutilSafeCall(cudaFree(cu_boundariesFactor34Odd));
	cutilSafeCall(cudaFree(cu_boundariesFactor56Odd));

	cutilSafeCall(cudaFree(cu_convergenceEvenTmp));
	cutilSafeCall(cudaFree(cu_convergenceEvenError));

	cutilSafeCall(cudaFree(cu_convergenceOddTmp));
	cutilSafeCall(cudaFree(cu_convergenceOddError));

	if (usingDivisor)
	{
	    cutilSafeCall(cudaFree(cu_even_divisor));
	    cutilSafeCall(cudaFree(cu_odd_divisor));
	}

	if (linear==0)
	{
		cutilSafeCall(cudaFree(cu_qmap1));
		cutilSafeCall(cudaFree(cu_qmap2));
		cutilSafeCall(cudaFree(cu_debmap1));
		cutilSafeCall(cudaFree(cu_debmap2));
		cutilSafeCall(cudaFree(cu_error));
	} 

	clock_gettime(CLOCK_REALTIME, &time1);
	profile.memoryRelease += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
}

void cu_poisson_runLaplace(const int nx, const int ny, const int nz, int color, const double c1, const double c2,
	CUDA_profile &profile)
{
	// Variables
	int sizeTotal;
	int sizeEven;
	int sizeOdd;
	
	dim3 threads(1, 1);
	dim3 grid(1, 1);
	
	int nxy = nx * ny;
	
	int sharedMemorySize;
	
	struct timespec time0, time1;
	
	// Configure kernel
	clock_gettime(CLOCK_REALTIME, &time0);
	threads.x = 16;
	threads.y = 16;
	grid.x = iDivUp(nx, threads.x - 2);
	grid.y = iDivUp(ny, threads.y - 2);
	sharedMemorySize = threads.x * threads.y * sizeof(double);
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
	
	// Calculate size
	clock_gettime(CLOCK_REALTIME, &time0);
	sizeTotal = nx * ny * nz;
	sizeOdd = sizeTotal / 2;
	sizeEven = sizeTotal - sizeOdd;
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
	
	// Bind data to textures
	clock_gettime(CLOCK_REALTIME, &time0);
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<doubleTex>();
	if (color == 0)
		cudaBindTexture(NULL, cu_phiTex, cu_phiOdd, channelDesc, sizeOdd * sizeof(double));
	else
		cudaBindTexture(NULL, cu_phiTex, cu_phiEven, channelDesc, sizeEven * sizeof(double));
	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.textureBind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
	
	// Launch kernel
	clock_gettime(CLOCK_REALTIME, &time0);
	if (color == 0)
	{
		cu_poisson_kernelLaplaceEven<<<grid, threads, sharedMemorySize>>>(nx, ny, nz, nxy,cu_phiEven,c1, c2);
	}
	else
	{
		cu_poisson_kernelLaplaceOdd<<<grid, threads, sharedMemorySize>>>(nx, ny, nz, nxy,cu_phiOdd,c1, c2);
	}
	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.kernelLaplace += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Unbind texture
	clock_gettime(CLOCK_REALTIME, &time0);
	cudaUnbindTexture(cu_phiTex);
	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.textureUnbind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
}


void cu_poisson_runBoltzmann(const int nx, const int ny, const int nz, int linear, int color, const double c2,
	CUDA_profile &profile)
{
	// Variables
	int sizeTotal;
	int sizeEven;
	int sizeOdd;
	
	dim3 threads(1, 1);
	dim3 grid(1, 1);
	int nxy = nx * ny;
	int sharedMemorySize;
	struct timespec time0, time1;
	
	// Configure kernel
	clock_gettime(CLOCK_REALTIME, &time0);
	threads.x = 16;
	threads.y = 16;
	grid.x = iDivUp(nx, threads.x - 2);
	grid.y = iDivUp(ny, threads.y - 2);
	sharedMemorySize = threads.x * threads.y * sizeof(double);
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
	
	// Calculate size
	clock_gettime(CLOCK_REALTIME, &time0);
	sizeTotal = nx * ny * nz;
	sizeOdd = sizeTotal / 2;
	sizeEven = sizeTotal - sizeOdd;
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
	
	// Bind data to textures
	clock_gettime(CLOCK_REALTIME, &time0);
	cudaChannelFormatDesc channelDescdouble = cudaCreateChannelDesc<doubleTex>();
	
	if (color == 0) 
	{
	    cudaBindTexture(NULL, cu_phiTex, cu_phiOdd, channelDescdouble, sizeOdd * sizeof(double));
	    cudaBindTexture(NULL, cu_divisorTex, cu_even_divisor, channelDescdouble, sizeOdd * sizeof(double));
	} 
	else 
	{
	    cudaBindTexture(NULL, cu_phiTex, cu_phiEven, channelDescdouble, sizeEven * sizeof(double));
	    cudaBindTexture(NULL, cu_divisorTex, cu_odd_divisor, channelDescdouble, sizeEven * sizeof(double));
	}

	if (!linear)
	{
	    if (color == 0)
			cudaBindTexture(NULL, cu_qmapTex, cu_qmap1, channelDescdouble, sizeEven * sizeof(double));
	    else
	    	cudaBindTexture(NULL, cu_qmapTex, cu_qmap2, channelDescdouble, sizeOdd * sizeof(double));	
	}
	
	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.textureBind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
	
	// Launch kernel
	clock_gettime(CLOCK_REALTIME, &time0);

	if (linear)
	{ 
	    if (color == 0)
	    {
			cu_poisson_kernelBoltzmannEven<<<grid, threads, sharedMemorySize>>>(nx,ny,nz,nxy,cu_phiEven,c2);
	    }
	    else
	    {
			cu_poisson_kernelBoltzmannOdd<<<grid, threads, sharedMemorySize>>>(nx,ny,nz,nxy,cu_phiOdd,c2);
	    }
	} 
	else
	{
	    if (color == 0)
	    {
			cu_poisson_kernelBoltzmannEven_nl<<<grid, threads, sharedMemorySize>>>(nx,ny,nz,nxy,cu_phiEven,c2);
	    }
	    else
	    {
			cu_poisson_kernelBoltzmannOdd_nl<<<grid, threads, sharedMemorySize>>>(nx,ny,nz,nxy,cu_phiOdd,c2);
	    }	    
	}

	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("Error running Boltzmann on %d: %s\n", color,cudaGetErrorString(err));

	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.kernelLaplace += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Unbind texture
	clock_gettime(CLOCK_REALTIME, &time0);
	cudaUnbindTexture(cu_phiTex);
	cudaUnbindTexture(cu_divisorTex);
	if (!linear)
		cudaUnbindTexture(cu_qmapTex);
	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.textureUnbind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

}

void cu_poisson_runCharges(const int nCharges, int color, CUDA_profile &profile)
{
	// Variables
	dim3 threads(1, 1);
	dim3 grid(1, 1);

	struct timespec time0, time1;

	// Configure kernel
	clock_gettime(CLOCK_REALTIME, &time0);
	threads.x = 256;
	threads.y = 1;
	grid.x = iDivUp(nCharges, threads.x);
	grid.y = 1;
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
	
	// Bind data to textures
	clock_gettime(CLOCK_REALTIME, &time0);

	cudaChannelFormatDesc channelDescPosition = cudaCreateChannelDesc<int>();
	cudaChannelFormatDesc channelDescCharges = cudaCreateChannelDesc<doubleTex>();

	if (color == 0)
	{
		cudaBindTexture(NULL, cu_chargesPositionTex, cu_chargesPositionEven, channelDescPosition, nCharges * sizeof(int));
		cudaBindTexture(NULL, cu_chargesTex, cu_chargesEven, channelDescCharges, nCharges * sizeof(double));
	}
	else
	{
		cudaBindTexture(NULL, cu_chargesPositionTex, cu_chargesPositionOdd, channelDescPosition, nCharges * sizeof(int));
		cudaBindTexture(NULL, cu_chargesTex, cu_chargesOdd, channelDescCharges, nCharges * sizeof(double));
	}

	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.textureBind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Launch kernel
	clock_gettime(CLOCK_REALTIME, &time0);
	if (color == 0)
		cu_poisson_kernelCharges<<<grid, threads>>>(cu_phiEven, nCharges);
	else
		cu_poisson_kernelCharges<<<grid, threads>>>(cu_phiOdd, nCharges);
	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.kernelCharges += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Unbind texture
	clock_gettime(CLOCK_REALTIME, &time0);
	cudaUnbindTexture(cu_chargesPositionTex);
	cudaUnbindTexture(cu_chargesTex);
	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.textureUnbind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("charges Error 1: %s\n", cudaGetErrorString(err));	

}

void cu_poisson_runDielectricBoundaries(const int nx, const int ny, const int nz,
	const int nBoundaries, int color, CUDA_profile &profile)
{
	// Variables
	int sizeTotal;
	int sizeEven;
	int sizeOdd;

	int offset_xn;
	int offset_xp;
	int offset_yn;
	int offset_yp;
	int offset_zn;
	int offset_zp;

	int nxy = nx * ny;

	dim3 threads(1, 1);
	dim3 grid(1, 1);

	struct timespec time0, time1;

	// Configure kernel
	clock_gettime(CLOCK_REALTIME, &time0);
	threads.x = 256;
	threads.y = 1;
	grid.x = iDivUp(nBoundaries, threads.x);
	grid.y = 1;
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Calculate size
	clock_gettime(CLOCK_REALTIME, &time0);

	sizeTotal = nx * ny * nz;
	sizeOdd = sizeTotal / 2;
	sizeEven = sizeTotal - sizeOdd;

	// Calculate offsets
	if (color == 0)
	{
		offset_xn = -1;
		offset_xp = 0;
		offset_yn = -nx / 2 - 1;
		offset_yp = nx / 2;
		offset_zn = -nxy / 2 - 1;
		offset_zp = nxy / 2;
	}
	else
	{
		offset_xn = 0;
		offset_xp = 1;
		offset_yn = -nx / 2;
		offset_yp = nx / 2 + 1;
		offset_zn = -nxy / 2;
		offset_zp = nxy / 2 + 1;
	}

	clock_gettime(CLOCK_REALTIME, &time1);
	profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Bind data to textures
	clock_gettime(CLOCK_REALTIME, &time0);

	cudaChannelFormatDesc channelDescPhi = cudaCreateChannelDesc<doubleTex>();
	cudaChannelFormatDesc channelDescPosition = cudaCreateChannelDesc<int>();
	cudaChannelFormatDesc channelDescFactor = cudaCreateChannelDesc<doubleTex2>();

	if (color == 0)
	{
		cutilSafeCall(cudaBindTexture(
			NULL, cu_phiTex, cu_phiOdd,
			channelDescPhi, sizeOdd * sizeof(double)));
		cutilSafeCall(cudaBindTexture(
			NULL, cu_boundariesPositionTex, cu_boundariesPositionEven,
			channelDescPosition, nBoundaries * sizeof(int)));
		cutilSafeCall(cudaBindTexture(
			NULL, cu_boundariesFactor12Tex, cu_boundariesFactor12Even,
			channelDescFactor, nBoundaries * sizeof(double) * 2));
		cutilSafeCall(cudaBindTexture(
			NULL, cu_boundariesFactor34Tex, cu_boundariesFactor34Even,
			channelDescFactor, nBoundaries * sizeof(double) * 2));
		cutilSafeCall(cudaBindTexture(
			NULL, cu_boundariesFactor56Tex, cu_boundariesFactor56Even,
			channelDescFactor, nBoundaries * sizeof(double) * 2));
	}
	else
	{
		cutilSafeCall(cudaBindTexture(
			NULL, cu_phiTex, cu_phiEven,
			channelDescPhi, sizeEven * sizeof(double)));
		cutilSafeCall(cudaBindTexture(
			NULL, cu_boundariesPositionTex, cu_boundariesPositionOdd,
			channelDescPosition, nBoundaries * sizeof(int)));
		cutilSafeCall(cudaBindTexture(
			NULL, cu_boundariesFactor12Tex, cu_boundariesFactor12Odd,
			channelDescFactor, nBoundaries * sizeof(double) * 2));
		cutilSafeCall(cudaBindTexture(
			NULL, cu_boundariesFactor34Tex, cu_boundariesFactor34Odd,
			channelDescFactor, nBoundaries * sizeof(double) * 2));
		cutilSafeCall(cudaBindTexture(
			NULL, cu_boundariesFactor56Tex, cu_boundariesFactor56Odd,
			channelDescFactor, nBoundaries * sizeof(double) * 2));
	}

	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.textureBind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Launch kernel
	clock_gettime(CLOCK_REALTIME, &time0);
	if (color == 0)
	{
		cu_poisson_kernelDielectricBoundaries<<<grid, threads>>>(
			nBoundaries, cu_phiEven,
			offset_xn, offset_xp,
			offset_yn, offset_yp,
			offset_zn, offset_zp);
	}
	else
	{
		cu_poisson_kernelDielectricBoundaries<<<grid, threads>>>(
			nBoundaries, cu_phiOdd,
			offset_xn, offset_xp,
			offset_yn, offset_yp,
			offset_zn, offset_zp);
	}

	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("Error running dielectric correction on %d: %s\n", color,cudaGetErrorString(err));

	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.kernelDielectricBoundaries += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	// Unbind texture
	clock_gettime(CLOCK_REALTIME, &time0);
	cudaUnbindTexture(cu_phiTex);
	cudaUnbindTexture(cu_boundariesPositionTex);
	cudaUnbindTexture(cu_boundariesFactor12Tex);
	cudaUnbindTexture(cu_boundariesFactor34Tex);
	cudaUnbindTexture(cu_boundariesFactor56Tex);
	cudaThreadSynchronize();
	clock_gettime(CLOCK_REALTIME, &time1);
	profile.textureUnbind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

}

void cu_poisson_runSaveLinearConvergenceCheck(const int nx, const int ny, const int nz,
	const int nPoints, const int color,
	CUDA_profile &profile)
{
    // Variables
    int sizeTotal;
    int sizeEven;
    int sizeOdd;

    dim3 threads(1, 1);
    dim3 grid(1, 1);

    struct timespec time0, time1;

    // Configure kernel
    clock_gettime(CLOCK_REALTIME, &time0);
    threads.x = 256;
    threads.y = 1;
    grid.x = iDivUp(nPoints, threads.x);
    grid.y = 1;
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Calculate size
    clock_gettime(CLOCK_REALTIME, &time0);

    sizeTotal = nx * ny * nz;
    sizeOdd = sizeTotal / 2;
    sizeEven = sizeTotal - sizeOdd;

    clock_gettime(CLOCK_REALTIME, &time1);
    profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Bind data to textures
    clock_gettime(CLOCK_REALTIME, &time0);

    cudaChannelFormatDesc channelDescPhi = cudaCreateChannelDesc<doubleTex>();
    cudaChannelFormatDesc channelDescPosition = cudaCreateChannelDesc<int>();

    if (color == 0)
    {
		cudaBindTexture(NULL, cu_phiTex, cu_phiEven, channelDescPhi, sizeEven * sizeof(double));
		
		cudaBindTexture(NULL, cu_chargesPositionTex, cu_chargesPositionEven, channelDescPosition, nPoints * sizeof(int));
    }
    else
    {
		cudaBindTexture(NULL, cu_phiTex, cu_phiOdd, channelDescPhi, sizeOdd * sizeof(double));
		
		cudaBindTexture(NULL, cu_chargesPositionTex, cu_chargesPositionOdd, channelDescPosition, nPoints * sizeof(int));
    }

    cudaThreadSynchronize();
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.textureBind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
	
    // Launch kernel
    clock_gettime(CLOCK_REALTIME, &time0);
    if (color == 0)
    {
		cu_poisson_kernelSaveConvergenceCheck<<<grid, threads>>>(nPoints, cu_convergenceEvenTmp);
    }
    else
    {
		cu_poisson_kernelSaveConvergenceCheck<<<grid, threads>>>(nPoints, cu_convergenceOddTmp);
    }
    
    cudaThreadSynchronize();
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.kernelSaveDielectricBoundaries += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Unbind texture
    clock_gettime(CLOCK_REALTIME, &time0);
    cudaUnbindTexture(cu_phiTex);
    cudaUnbindTexture(cu_chargesPositionTex);
    cudaThreadSynchronize();
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.textureUnbind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
}

void cu_poisson_initComparePhi(const int sizeEven, const int sizeOdd, CUDA_profile &profile)
{
    struct timespec time0, time1;
    

//	Memory allocation
    clock_gettime(CLOCK_REALTIME, &time0);

    cutilSafeCall(cudaMalloc((void**)&cu_phiEven2, sizeEven*sizeof(double)));
    cutilSafeCall(cudaMalloc((void**)&cu_phiOdd2, sizeOdd*sizeof(double)));
	cutilSafeCall(cudaMalloc((void**)&cu_error, 256*sizeof(double)));
	
	clock_gettime(CLOCK_REALTIME, &time1);
	
    profile.memoryAllocation += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("init Compare Error: %s\n", cudaGetErrorString(err));

	return;
}

void cu_poisson_runSavePhimap(const int ntp, int color, CUDA_profile &profile)
{
    struct timespec time0, time1;

    // Configure kernel
    clock_gettime(CLOCK_REALTIME, &time0);

	dim3 threads(1, 1);
    dim3 grid(1, 1);

    threads.x = 256;
    threads.y = 1;
    
    grid.x = 256;
    grid.y = 1;

	int perThread = iDivUp(ntp, threads.x * grid.x);

    clock_gettime(CLOCK_REALTIME, &time1);
    profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
    
    // Bind data to textures
    clock_gettime(CLOCK_REALTIME, &time0);
    cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<doubleTex>();

	if (color == 0)
		cudaBindTexture(NULL, cu_phiTex, cu_phiEven, channelDesc, ntp*sizeof(double));
	else
		cudaBindTexture(NULL, cu_phiTex, cu_phiOdd, channelDesc, ntp*sizeof(double));

    cudaThreadSynchronize();
	
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.textureBind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Launch kernel
    clock_gettime(CLOCK_REALTIME, &time0);
    if (color == 0)
    {
		cu_poisson_kernelSavePhi<<<grid, threads>>>(ntp, cu_phiEven2, perThread);
    }
    else
    {
		cu_poisson_kernelSavePhi<<<grid, threads>>>(ntp, cu_phiOdd2, perThread);
    }
    
    cudaThreadSynchronize();
	
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.kernelCompareDielectricBoundaries += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Unbind texture
    clock_gettime(CLOCK_REALTIME, &time0);
    cudaUnbindTexture(cu_phiTex);
    cudaThreadSynchronize();
	
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.textureUnbind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);    

 	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("save phimap Error: %s\n", cudaGetErrorString(err));

}

double cu_poisson_runComparePhi(int ntp, int color, CUDA_profile &profile)
{	

	double maxError;
	struct timespec time0, time1;
	clock_gettime(CLOCK_REALTIME, &time0);

	dim3 threads(1, 1);
    dim3 grid(1, 1);

    threads.x = 256;
    threads.y = 1;
    
    grid.x = 256;
    grid.y = 1;

	int perThread = iDivUp(ntp, threads.x * grid.x);

    int sharedMemorySize = threads.x * sizeof(double);
	
	double *globalErrors;
	globalErrors = new double[grid.x];
	
	clock_gettime(CLOCK_REALTIME, &time1);
    profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
	clock_gettime(CLOCK_REALTIME, &time0);
	
    cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<doubleTex>();

	if (color == 0) {
		cutilSafeCall(cudaBindTexture(NULL, cu_phiTex, cu_phiEven, channelDesc, ntp*sizeof(double)));
		cutilSafeCall(cudaBindTexture(NULL, cu_phiTex2, cu_phiEven2, channelDesc, ntp*sizeof(double)));
	} else {
		cutilSafeCall(cudaBindTexture(NULL, cu_phiTex, cu_phiOdd, channelDesc, ntp*sizeof(double)));
		cutilSafeCall(cudaBindTexture(NULL, cu_phiTex2, cu_phiOdd2, channelDesc, ntp*sizeof(double)));}

    cudaThreadSynchronize();

    clock_gettime(CLOCK_REALTIME, &time1);
    profile.textureBind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
	clock_gettime(CLOCK_REALTIME, &time0);

	cu_poisson_kernelComparePhi <<<grid, threads, sharedMemorySize>>> (ntp, cu_error, perThread);

	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("Error checking convergence: %s\n", cudaGetErrorString(err));

	cudaThreadSynchronize();
	cudaUnbindTexture(cu_phiTex);
	cudaUnbindTexture(cu_phiTex2);
	cudaThreadSynchronize();


    clock_gettime(CLOCK_REALTIME, &time1);
    profile.kernelCompareDielectricBoundaries += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    clock_gettime(CLOCK_REALTIME, &time0);

	cutilSafeCall(cudaMemcpy(globalErrors, cu_error, grid.x*sizeof(double), cudaMemcpyDeviceToHost));
	
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.device2host += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Get max error
	maxError = 0.0;
	
    clock_gettime(CLOCK_REALTIME, &time0);
    for (int ii = 0; ii < grid.x; ii++)
    {
    	// printf("%f ",globalErrors[ii]);
		if (globalErrors[ii] > maxError)
			maxError = globalErrors[ii];
    }
    // printf("\n");

    delete [] globalErrors;
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
    
	return maxError;
    
}

double cu_poisson_runLinearCheckConvergence(const int nx, const int ny, const int nz, const int nPoints, int color,
	CUDA_profile &profile)
{
    // Variables
    int sizeTotal;
    int sizeEven;
    int sizeOdd;
	
    dim3 threads(1, 1);
    dim3 grid(1, 1);
	
    double *boundaryErrors;
    double maxError = 0.0;
	
    struct timespec time0, time1;

    // Configure kernel
    clock_gettime(CLOCK_REALTIME, &time0);
    threads.x = 256;
    threads.y = 1;
    grid.x = iDivUp(nPoints, threads.x);
    grid.y = 1;
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Calculate size
    clock_gettime(CLOCK_REALTIME, &time0);

    sizeTotal = nx * ny * nz;
    sizeOdd = sizeTotal / 2;
    sizeEven = sizeTotal - sizeOdd;

    // Init boundary errors
    boundaryErrors = new double[nPoints];

    clock_gettime(CLOCK_REALTIME, &time1);
    profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Bind data to textures
    clock_gettime(CLOCK_REALTIME, &time0);

    cudaChannelFormatDesc channelDescPhi = cudaCreateChannelDesc<doubleTex>();
    cudaChannelFormatDesc channelDescPosition = cudaCreateChannelDesc<int>();

    if (color == 0)
    {
		cudaBindTexture(NULL, cu_phiTex, cu_phiEven, channelDescPhi, sizeEven * sizeof(double));
		cudaBindTexture(NULL, cu_chargesPositionTex, cu_chargesPositionEven, channelDescPosition, nPoints * sizeof(int));
    }
    else
    {
		cudaBindTexture(NULL, cu_phiTex, cu_phiOdd, channelDescPhi, sizeOdd * sizeof(double));
		cudaBindTexture(NULL, cu_chargesPositionTex, cu_chargesPositionOdd, channelDescPosition, nPoints * sizeof(int));
    }

    cudaThreadSynchronize();
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.textureBind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Launch kernel
    clock_gettime(CLOCK_REALTIME, &time0);
    if (color == 0)
    {
		cu_poisson_kernelLinearConvergenceCheck<<<grid, threads>>>(nPoints, cu_convergenceEvenTmp, cu_convergenceEvenError);
    }
    else
    {
		cu_poisson_kernelLinearConvergenceCheck<<<grid, threads>>>(nPoints, cu_convergenceOddTmp, cu_convergenceOddError);
    }
    
    cudaThreadSynchronize();
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.kernelCompareDielectricBoundaries += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    cudaUnbindTexture(cu_phiTex);
    cudaUnbindTexture(cu_chargesPositionTex);
    cudaThreadSynchronize();
    
    // Get boundary errors
    clock_gettime(CLOCK_REALTIME, &time0);
    if (color == 0)
		cutilSafeCall(cudaMemcpy(boundaryErrors, cu_convergenceEvenError, nPoints * sizeof(double), cudaMemcpyDeviceToHost));
    else
		cutilSafeCall(cudaMemcpy(boundaryErrors, cu_convergenceOddError, nPoints * sizeof(double), cudaMemcpyDeviceToHost));
    
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.device2host += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Get max error
    clock_gettime(CLOCK_REALTIME, &time0);
    for (int i = 0; i < nPoints; i++)
    {
		if (maxError < boundaryErrors[i])
		    maxError = boundaryErrors[i];
    }
    
    delete [] boundaryErrors;
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("Error checking linear convergence: %s\n", cudaGetErrorString(err));

    return maxError;
}


void cu_poisson_initQmap(const int nx, const int ny, const int nz,
    int *debmap1, int *debmap2, double *qmap1, double *qmap2, CUDA_profile &profile)
{
    // Variables
    int sizeTotal;
    int sizeEven;
    int sizeOdd;
    struct timespec time0, time1;

    // Calculate size
    clock_gettime(CLOCK_REALTIME, &time0);
    sizeTotal = nx * ny * nz;
    sizeOdd = (sizeTotal-1) / 2;
    sizeEven = sizeTotal - sizeOdd;
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.others += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Memory allocation
    clock_gettime(CLOCK_REALTIME, &time0);
    cutilSafeCall(cudaMalloc((void**)&cu_qmap1, sizeEven * sizeof(double)));
    cutilSafeCall(cudaMalloc((void**)&cu_qmap2, sizeOdd * sizeof(double)));
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.memoryAllocation += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Copy data
    clock_gettime(CLOCK_REALTIME, &time0);
    cutilSafeCall(cudaMemcpy(cu_qmap1, qmap1, sizeEven * sizeof(double), cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(cu_qmap2, qmap2, sizeOdd * sizeof(double), cudaMemcpyHostToDevice));
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.host2device += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    clock_gettime(CLOCK_REALTIME, &time0);
    cutilSafeCall(cudaMalloc((void**)&cu_debmap1, sizeEven * sizeof(int)));
    cutilSafeCall(cudaMalloc((void**)&cu_debmap2, sizeOdd * sizeof(int)));

    cutilSafeCall(cudaMemcpy(cu_debmap1, debmap1, sizeEven * sizeof(int), cudaMemcpyHostToDevice));
    cutilSafeCall(cudaMemcpy(cu_debmap2, debmap2, sizeOdd * sizeof(int), cudaMemcpyHostToDevice));
    profile.memoryAllocation += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);
    
    cu_debTex.normalized = false;
    cu_debTex.filterMode = cudaFilterModePoint;
    cu_debTex.addressMode[0] = cudaAddressModeClamp;
    
    cu_qmapTex.normalized = false;
    cu_qmapTex.filterMode = cudaFilterModePoint;
    cu_qmapTex.addressMode[0] = cudaAddressModeClamp;

	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("Error at qmap initialization: %s\n", cudaGetErrorString(err));

}

void cu_poisson_updatePointers(
        int nx, int ny, int nz, double omcomp, 
		int numberEvenCharges, int numberOddCharges,
		int nEvenDielectricBoundaries, int nOddDielectricBoundaries,
		CUDA_profile &profile)
{

    struct timespec time0, time1;
    clock_gettime(CLOCK_REALTIME, &time0);
    
    int nt = nx * ny * nz;
    int neven, nodd;
    neven = (nt+1)/2;
    nodd = neven-1;
    dim3 threads(1, 1);
    dim3 grid(1, 1);

    threads.x = 256;
    threads.y = 1;
    grid.y = 1;
	grid.x = 256;
		
	int perThread;	
    /** ****************
    *  UPDATE OF THE DIVISOR */
    perThread = iDivUp(neven, threads.x*grid.x);
    cudaThreadSynchronize();
    cu_poisson_kernelUpdateDivisor<<<grid, threads>>>(neven, cu_even_divisor, omcomp, perThread);

    perThread = iDivUp(nodd, threads.x*grid.x);
    cudaThreadSynchronize();
    cu_poisson_kernelUpdateDivisor<<<grid, threads>>>(nodd, cu_odd_divisor, omcomp, perThread);
    cudaThreadSynchronize();
    
    /** ****************
    *  UPDATE OF THE CHARGES */
    perThread = iDivUp(numberEvenCharges, threads.x*grid.x);
    cu_poisson_kernelUpdateCharges<<<grid, threads>>>(cu_chargesEven, numberEvenCharges, omcomp, perThread);
    cudaThreadSynchronize();
    perThread = iDivUp(numberOddCharges, threads.x*grid.x);
    cu_poisson_kernelUpdateCharges<<<grid, threads>>>(cu_chargesOdd, numberOddCharges, omcomp, perThread);
    cudaThreadSynchronize();

    /** ****************
    *  UPDATE OF THE DIELECTRIC BOUNDARIES */
    perThread = iDivUp(2*nEvenDielectricBoundaries, threads.x*grid.x);
    cu_poisson_kernelUpdateDielectrics<<<grid, threads>>>(nEvenDielectricBoundaries, omcomp,
		cu_boundariesFactor12Even, cu_boundariesFactor34Even, cu_boundariesFactor56Even, perThread);
    cudaThreadSynchronize();
    perThread = iDivUp(2*nOddDielectricBoundaries, threads.x*grid.x);
    cu_poisson_kernelUpdateDielectrics<<<grid, threads>>>(nOddDielectricBoundaries, omcomp,
		cu_boundariesFactor12Odd, cu_boundariesFactor34Odd, cu_boundariesFactor56Odd, perThread);
    cudaThreadSynchronize();

    clock_gettime(CLOCK_REALTIME, &time1);
    profile.updatePointers += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("Error updating pointers: %s\n", cudaGetErrorString(err));
}


void cu_poisson_updateQmap(int nt, int parity, double fac1, double chi2, double chi3, 
	double chi4, double chi5, CUDA_profile &profile)
{
    struct timespec time0, time1;
    clock_gettime(CLOCK_REALTIME, &time0);
	
    dim3 threads(1,1);
    dim3 grid(1,1);
    threads.x = 256;
    threads.y = 1;
    grid.y = 1;
    grid.x = 256;
    int perThread = iDivUp(nt, threads.x*grid.x);
    
    cudaChannelFormatDesc channelDescdouble = cudaCreateChannelDesc<doubleTex>();
    cudaChannelFormatDesc channelDescInt = cudaCreateChannelDesc<int>();
    
    if (parity == 0)
    {
		cutilSafeCall(cudaBindTexture(NULL, cu_phiTex, cu_phiEven, channelDescdouble, nt * sizeof(double)));
		cutilSafeCall(cudaBindTexture(NULL, cu_debTex, cu_debmap1, channelDescInt, nt * sizeof(int)));
    }
    else
    {
		cutilSafeCall(cudaBindTexture(NULL, cu_phiTex, cu_phiOdd, channelDescdouble, nt * sizeof(double)));
		cutilSafeCall(cudaBindTexture(NULL, cu_debTex, cu_debmap2, channelDescInt, nt * sizeof(int)));
    }

    cudaThreadSynchronize();
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.textureBind += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

    // Launch kernel
    clock_gettime(CLOCK_REALTIME, &time0);

    if (parity == 0)
    {
		cu_poisson_kernelUpdateQmap <<<grid, threads>>> (nt, fac1, chi2, chi3, chi4, chi5, cu_qmap1, perThread);
    }
    else
    {
		cu_poisson_kernelUpdateQmap <<<grid, threads>>> (nt, fac1, chi2, chi3, chi4, chi5, cu_qmap2, perThread);
    }


    cudaThreadSynchronize();
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.updatePointers += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);


    cudaUnbindTexture(cu_phiTex);
    cudaUnbindTexture(cu_debTex);

    cudaThreadSynchronize();
    clock_gettime(CLOCK_REALTIME, &time1);
    profile.updatePointers += (double)(time1.tv_sec - time0.tv_sec) + 1E-9 * (double)(time1.tv_nsec - time0.tv_nsec);

	cudaError_t err = cudaGetLastError();
	if (err != cudaSuccess) 
	    printf("Error updating qmap: %s\n", cudaGetErrorString(err));
}

void cu_getDeviceInfo()
{
	cuda_getDeviceInfo();
}

void cu_init(int device)
{
	cuda_init(device);
}

void cu_finish()
{
	cuda_finish();
}
