#include <nlfdsolver.h>
	
void nlfdsolver(double salt, int * __restrict__ idpos, double * dbcorrection, 
	double * __restrict__ even_divisor, double * __restrict__ odd_divisor, int * __restrict__ charge_index, 
	double * __restrict__ qval, double * __restrict__ even_phi, double * __restrict__ odd_phi,
	int number_dielectrics1, int number_dielectrics2, int number_charges1, int number_charges2, 
	double omega, int nx, int ny, int nz, double max_error, int max_iterations, string calculation_mode,
	bool *inoutPointer, bool imanual, int nnit, int *ival, int *ival2, double *conc, double *qmap1,
	double *qmap2, double qfact, double epsout, double deblen, double scale, double relpar)
{


    struct timespec time0, time1;

    double timeInit = 0.0;
    double timeLaplace = 0.0;
    double timeCharges = 0.0;
    double timeBoundaries = 0.0;
    double timeDielectricBoundaries = 0.0;
    double timeOthers = 0.0;
    double timeTotal = 0.0;
    double timeIteration = 0.0;
    double timeUpdate = 0.0;
    double timeCheckConvergence = 0.0;
    int totalIterations = 1;

    printf("Starting nonlinear numerical solver\n");
    
    int ii, nlinear=0;
    double loop_error, f1;
    
    int iter_convergence_check = 10;
    const int ntp = (nx*ny*nz+1)/2;
	const int nop = ntp-1;
	const int nep = ntp;

    int icountplus;
    double factor;
    bool ichangeom, inewfirst, istop, inew;
    double om1, fraction = 0.0;
    int itnum = 0, ires = 0;
    double conv[4];
    double derprec, der;
    int icont = 0;
    
    double omcomp = 0, relparprev;
    double debfct;
    double cost;
    double fac1;
    
    double z1p, z2p, z1m, z2m, cz1, cz2;
    double chi2, chi3, chi4, chi5;
    
    conv[0] = 0.0;
    conv[1] = 0.0;
    conv[2] = 0.0;
    conv[3] = 0.0;
    relparprev = omega;
    der = 0.0;
    
    icountplus=0; 
    factor=1.0;
    ichangeom = true;
    inewfirst = false;
    istop = true;
    inew = true;
    
    z1p = ival[0];
    z2p = ival2[0];
    z1m = ival[1];
    z2m = ival2[1];
    
    cz1 = conc[0]*z1m;
    cz2 = conc[1]*z2m;
    
    f1=(cz1*z1p*(z1p+z1m)+cz2*z2p*(z2p+z2m))/2.0;
    
    chi2 = (cz1*z1p*pow(z1p,2.0)-cz1*z1m*pow(z1m,2.0)+cz2*z2p*pow(z2p,2.0)-cz2*z2m*pow(z2m,2.0))/2.0;
    chi3 = -(cz1*z1p*pow(z1p,3.0)+cz1*z1m*pow(z1m,3.0)+cz2*z2p*pow(z2p,3.0)+cz2*z2m*pow(z2m,3.0))/6.0;
    chi4 = (cz1*z1p*pow(z1p,4.0)-cz1*z1m*pow(z1m,4.0)+cz2*z2p*pow(z2p,4.0)-cz2*z2m*pow(z2m,4.0))/24.0;
    chi5 = -(cz1*z1p*pow(z1p,5.0)+cz1*z1m*pow(z1m,5.0)+cz2*z2p*pow(z2p,5.0)+cz2*z2m*pow(z2m,5.0))/120.0;
    
    scale = 1.0/scale;
    debfct = epsout/pow(deblen*scale,2.0);
    cost = debfct/(2.0*f1*epsout);

    if (imanual) {
		ichangeom = true;
    } else {
		ichangeom = false;}
    
    inewfirst = false; 
    istop = true;
    inew = true;
    
    for (ii=0; ii<ntp; ii++)
    	qmap1[ii] = 0.0;

    for (ii=0; ii<ntp; ii++)
    	qmap2[ii] = 0.0;
    
	//////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////
	//////////////////////////////////////////////////////////////////////////

    //time0 = omp_get_wtime();

    bool usingDivisor;
    if (salt==0.0)
    	usingDivisor = false;
    else
    	usingDivisor = true;


    om1 = 1-omega;
    

    int *debmap1 = new int[ntp];
    int *debmap2 = new int[ntp];
    
    for (ii=0; ii<ntp; ii++)
    {
		debmap1[ii] = inoutPointer[2*ii] ? 1 : 0;
		debmap2[ii] = inoutPointer[2*ii+1] ? 1 : 0;
    }

//     int current_index;

    printf("Starting to iterate ********** \n");

	initProfiler(time0);

	Float maxErrorEven, maxErrorOdd;

    CUDA_profile cudaProfile = {0.0};

    double *even_dielectric_boundary_correction = dbcorrection;
    double *odd_dielectric_boundary_correction = dbcorrection + number_dielectrics1*6;
    int *even_dielectric_boundaries = idpos;
    int *odd_dielectric_boundaries = idpos + number_dielectrics1;

    int *even_charges_index = charge_index;
    int *odd_charges_index = charge_index + number_charges1;
    
    double  * __restrict__ even_bb = qval;
    double  * __restrict__ odd_bb = qval + number_charges1;

    const int number_even_charges = number_charges1;
    const int number_odd_charges = number_charges2 - number_charges1;
    const int number_even_dielectric_boundaries = number_dielectrics1;
    const int number_odd_dielectric_boundaries = number_dielectrics2 - number_dielectrics1;

	printf("initializating CUDA\n");

	cu_init(0);

	cu_getDeviceInfo();

	// Init memory
	cu_poisson_initPhi(nx, ny, nz, cudaProfile);
	
	// Copy initial data
	cu_poisson_copyPhiToGPU(nx, ny, nz, even_phi, odd_phi, cudaProfile);
	
	// Copy charges
	cu_poisson_initCopyCharges(nx, ny, nz, 0, number_even_charges, even_charges_index, even_bb,
		number_odd_charges, odd_charges_index, odd_bb, cudaProfile);

	// Copy dielectric boundaries
	cu_poisson_initCopyDielectricBoundaries(nx, ny, nz,
		number_even_dielectric_boundaries, even_dielectric_boundaries, even_dielectric_boundary_correction,
		number_odd_dielectric_boundaries, odd_dielectric_boundaries, odd_dielectric_boundary_correction,
		0, cudaProfile);
		
	cu_poisson_initComparePhi(ntp,ntp-1,cudaProfile);

	cu_poisson_initCopyDivisor(nep,nop, even_divisor, odd_divisor, cudaProfile);

	cu_poisson_initQmap(nx, ny, nz, debmap1, debmap2, qmap1, qmap2, cudaProfile);

    profiler(time0,time1,timeInit);

    int nn = 1;
    do
    {
		
		ifstream my_file("abort.now");
		if (my_file.good())
		{
		    printf("Found abort.now file. Aborting\n");
		    my_file.close();
		    break;
		}

		if (nn%iter_convergence_check == 0)
		{

			cu_poisson_runSavePhimap(ntp, 0, cudaProfile);
			cu_poisson_runSavePhimap(ntp-1, 1, cudaProfile);
		}

		profiler(time0,time1,timeCheckConvergence);

	    // iterate over odd grid points to solve laplace only
	    cu_poisson_runBoltzmann(nx, ny, nz, 0, 1, om1, cudaProfile);

		profiler(time0,time1,timeLaplace);

	    cu_poisson_runDielectricBoundaries(nx, ny, nz, number_odd_dielectric_boundaries, 1, cudaProfile);		

		profiler(time0,time1,timeDielectricBoundaries);

	    cu_poisson_runCharges(number_odd_charges, 1, cudaProfile);

		profiler(time0,time1,timeCharges);	

        cu_poisson_runBoltzmann(nx, ny, nz, 0, 0, om1, cudaProfile);

		profiler(time0,time1,timeLaplace);

	    cu_poisson_runDielectricBoundaries(nx, ny, nz, number_even_dielectric_boundaries, 0, cudaProfile);

		profiler(time0,time1,timeDielectricBoundaries);

	    cu_poisson_runCharges(number_even_charges, 0, cudaProfile);

		profiler(time0,time1,timeCharges);

		if (nn%iter_convergence_check == 0) /* test convergence */
		{

			maxErrorEven = cu_poisson_runComparePhi(ntp, 0, cudaProfile);
			maxErrorOdd = cu_poisson_runComparePhi(ntp-1, 1, cudaProfile);			
			loop_error = max(maxErrorEven, maxErrorOdd);

		    profiler(time0,time1,timeCheckConvergence);

		    conv[2] = conv[1];
		    conv[1] = conv[0];
		    conv[0] = loop_error; 

		    if ( (loop_error < max_error) && (itnum > 22) )
		    {
			    ires = 1;
		    }

		    inew = inew || inewfirst;
		    
		    if (itnum==0)
		    {
			    printf("Iteration %d error %f\n",nn,loop_error);
			
			    istop = loop_error > 0.22 ? false : true;
			    
			    inewfirst = ((nn > max_iterations-max_iterations%iter_convergence_check) && istop && itnum==0);

			    if (!imanual && inewfirst && qfact > 3.2)
			    {
					factor = exp(-qfact*2.1)+1.E-6;
					ichangeom = true;
			    }
		    }
		    else
		    {			
				if (!imanual)
				{
				    derprec=der;
				    der=(conv[0]-conv[1])/conv[0];
				    
				    if (loop_error < 1.E-6)
				    {
						factor = 1.2;
						ichangeom = true;
				    }
				    else
				    {
						if ((der > 0.0) && (!inew))
						{
						    icountplus++;
						    factor = factor*pow(1.0-der,0.99);
						    if (der > 0.55)
						    {
								ichangeom = true;
								
								if (der >= 1.0)
								    factor = 1.E-5;
							
						    }
						    
						    if (der > 0.35 && conv[0] > 0.1)
						    {
								ichangeom = true;
								factor = pow(factor*0.05/conv[0],4.0);
						    }
						}
						
						if (der > 0.0 && conv[0] > 0.1 && inew && !inewfirst)
						{
						    ichangeom = true;
						    factor = min(factor*0.05/conv[0],pow(factor*(1.0-der),0.86));
						}
						
						if (der <= 0.0)
						{
						    icountplus = 0;
						    factor = 1.0;
						    if (itnum > 24 && itnum < 24+0.75*(nnit-24) && loop_error < 0.03 && derprec < 0.0)
						    {
								if (der > -0.2 && derprec > -0.2)
								{
									printf("Trying to speed up the convergence process \n");
									factor = 1.1;
									ichangeom = true;
									if (relpar < 0.2 && der > -0.05 && derprec > -0.5)
										factor = 1.0-45.226*(relpar-0.2);
								}
						    }
						}
						
						if (icountplus >= 2) 
						    ichangeom = true;
				    }
				    inewfirst = false;
				}	

				printf("Iteration %d error %E non-linear fraction %f factor %f\n",itnum,loop_error,fraction,factor);

		    }

		    profiler(time0,time1,timeOthers);

		}
		
		nn++;
		if ( ((nn <= max_iterations) || !istop) && (ires == 0)) 
		{
		    continue;
		}
		
		if (nnit > 0 && itnum == 0){
		    printf("Now for the non-linear iterations \n");
            nlinear = nn;}
		
		iter_convergence_check=10;
		max_iterations=10;
		itnum++;
		
		if ((itnum > nnit) || ((ires==1) && (!inew)))
		{
		    printf("Convergence obtained\n");
		    break;
		}
		
		nn = 1;
		
		inew = ichangeom;

		profiler(time0,time1,timeOthers);

		if (ichangeom)
		{   
		    relpar *= factor;

		    if (relpar < 1.E-4)
		    {
				printf("Estimation %f 1.E-4 prefered. \n",relpar);
				relpar = 1.E-4;
		    }
		    
		    factor = 1.0;
		    
		    ichangeom = false;
		    icountplus = 0;
		    omcomp = relpar/relparprev;
		    relparprev = relpar;
		    om1 = 1.0-relpar;

			profiler(time0,time1,timeOthers);
    
			cu_poisson_updatePointers(nx, ny, nz, omcomp, number_even_charges, number_odd_charges,
	            number_even_dielectric_boundaries, number_odd_dielectric_boundaries, cudaProfile);

			icont++;

		    profiler(time0,time1,timeUpdate);

		}
		
		fraction = fraction + 0.05;
		if(fraction > 1.0)
		{
			fraction = 1.0;
		} 
		    
		
		fac1 = fraction*cost;


	    cu_poisson_updateQmap(ntp,0,fac1,chi2,chi3,chi4,chi5,cudaProfile);
	    cu_poisson_updateQmap(ntp-1,1,fac1,chi2,chi3,chi4,chi5,cudaProfile);


	    profiler(time0,time1,timeUpdate);

    } while(true);

	cu_poisson_copyPhiFromGPU(nx, ny, nz,even_phi, odd_phi,cudaProfile);

    if (nn == max_iterations-1)
    {
		printf("  convergence was never obtained\n");
    }


    totalIterations = nlinear+itnum;
    printf("Total iterations = %d\n",totalIterations);

    timeTotal =
	timeInit + timeLaplace +
	timeCharges + timeBoundaries +
	timeDielectricBoundaries + timeOthers + timeUpdate;
    timeIteration =
	(timeLaplace + timeCharges +
	timeBoundaries + timeDielectricBoundaries +
	timeOthers) / totalIterations;

    printf("  General time (seconds):\n");
    printf("    Init                %f9\n",timeInit);
    if (usingDivisor)
    {
    	printf("    Boltzmann             %f9\n",timeLaplace);
    }
    else
    {
    	printf("    Laplace             %f9\n",timeLaplace);
    }
    printf("    Adjust charges      %f9\n",timeCharges);
    printf("    Sum boundaries      %f9\n",timeBoundaries);
    printf("    Adjust diel. bound. %f9\n",timeDielectricBoundaries);
    printf("    Non-linear update   %f9\n",timeUpdate);
    printf("    Others              %f9\n",timeOthers);
    printf("    Total               %f9\n",timeTotal);
    printf("    Per iteration       %f9\n",timeIteration);


	cu_poisson_destroyMemory(usingDivisor, 1, cudaProfile);

	timeTotal =
		cudaProfile.memoryAllocation +
		cudaProfile.memoryRelease +
		cudaProfile.textureBind +
		cudaProfile.textureUnbind +
		cudaProfile.host2device +
		cudaProfile.device2host +
		cudaProfile.device2device +
		cudaProfile.prepareData +
		cudaProfile.kernelLaplace +
		cudaProfile.kernelCharges +
		cudaProfile.kernelBoundaries +
		cudaProfile.kernelDielectricBoundaries +
		cudaProfile.kernelSaveDielectricBoundaries +
		cudaProfile.kernelCompareDielectricBoundaries +
		cudaProfile.others;

	printf("  CUDA time (seconds):\n");
	printf("    Memory Allocation   %f9\n",cudaProfile.memoryAllocation);
	printf("    Memory Release      %f9\n",cudaProfile.memoryRelease);
	printf("    Texture bind        %f9\n",cudaProfile.textureBind);
	printf("    Texture unbind      %f9\n",cudaProfile.textureUnbind);
	printf("    Host to Device      %f9\n",cudaProfile.host2device);
	printf("    Device to Host      %f9\n",cudaProfile.device2host);
	printf("    Device to Device    %f9\n",cudaProfile.device2device);
	printf("    Prepare Data        %f9\n",cudaProfile.prepareData);
	printf("    Kernel Laplace      %f9\n",cudaProfile.kernelLaplace);
	printf("    Kernel Charges      %f9\n",cudaProfile.kernelCharges);
	printf("    Kernel Boundaries   %f9\n",cudaProfile.kernelBoundaries);
	printf("    Kernel Diel. Bound. %f9\n",cudaProfile.kernelDielectricBoundaries);
	printf("    Kernel Save Bound.  %f9\n",cudaProfile.kernelSaveDielectricBoundaries);
	printf("    Kernel Comp. Bound. %f9\n",cudaProfile.kernelCompareDielectricBoundaries);
	printf("    Others              %f9\n",cudaProfile.others);
	printf("    Total               %f9\n",timeTotal);


	cu_finish();

    return;
}

