import os,sys,inspect,pickle,shutil,subprocess
import platform

############################## include path to find diff_match_patch_module  ##############################
cmd_folder = os.path.realpath(os.path.abspath(os.path.split(inspect.getfile( inspect.currentframe() ))[0]))
if cmd_folder not in sys.path:
  sys.path.insert(0, cmd_folder)

cmd_subfolder = os.path.realpath(os.path.abspath(os.path.join(os.path.split(inspect.getfile( inspect.currentframe() ))[0],"./diff_match_patch/python2")))
if cmd_subfolder not in sys.path:
  sys.path.insert(0, cmd_subfolder)
############################################################################################################

import diff_match_patch as dmp_module
# Force a module reload.  Allows one to edit the DMP module and rerun the tests
# without leaving the Python interpreter.
reload(dmp_module)

def main():
    
  dmp = dmp_module.diff_match_patch()  
  
  print '------------------'
  print 'DelPhi patcher 1.0'
  print '------------------'
  print ''
  print 'This procedure will guide you through patching DelPhi'
  print 'and compiling the Solver and NanoShaper modules.'
  print 'If you want to compile/install Solver/NanoShaper modules'
  print 'they must be present under the DelPhi folder'
  
  print ''  
  raw_input('Press any key to continue...')
  print ''
  
  #list of files to be patched
  filesList = ['qdiff4v.f','qinttot.f','pointer.h','qlog.h',
  'qdiffpar5.h','wrt4.f','distrTOpoint.f','crgarr.f','getatm2.f',
  'wrtphi.f','wrtsit4.f']

  newFiles = ['surfacemodule.cpp','solvermodule.cpp','nlsolvermodule.cpp']
  
  NSmodule = False
  SolverModule = False
  patchesDir = './patches/'
  originalSourcesDir = './src/'
  solverDir = './SolverCuda/'
  nsDir = './NanoShaper/'
  
  if(os.path.isdir(originalSourcesDir)==False):
    print "<<ERROR>> Cannot find sources folder"
    quit()
  
  if(os.path.isdir(patchesDir)==False):
    print "<<ERROR>> Cannot find patches folder"
    quit()

  if(os.path.isdir(solverDir)==True):
    print "<<INFO>> Solver module found!"
    SolverModule = True

  if(os.path.isdir(nsDir)==True):
    print "<<INFO>> NanoShaper module found!"
    NSmodule = True

  print ''
  print 'Start patching...'
  print ''
  
  for file in filesList:

    patchFile = open(patchesDir+file+'.patch','rb')
    bfile = file+'.back_off'
    boff = originalSourcesDir+bfile
    
    if os.path.isfile(boff):
      print 'Recovering original file ->',file
      shutil.copy2(boff,originalSourcesDir+file)
    else:
      print 'Backing-off file ->',file,' into ->',bfile
      shutil.copy2(originalSourcesDir+file,boff)
    
    print 'Patching file ->',file
    patches = pickle.load(patchFile)
    patchFile.close()
    src1 = open(originalSourcesDir+file, 'r')
    originalSource = src1.read()
    patched_file = dmp.patch_apply(patches,originalSource)[0]
    tt = open(originalSourcesDir+file,'w')
    tt.write(patched_file)
    tt.close()
  
  print ''
  
  for file in newFiles:
    print 'Adding new source ->',file
    shutil.copy2(patchesDir+file,originalSourcesDir+file)

  print ''
  print 'Updating makefile...'
  
  shutil.copy2('./patches/GCC/makefile','./')
  shutil.copy2('./patches/GCC/genericLinux.mk','./src')
  
  print 'Compiling patched DelPhi...'  
  subprocess.call('make',shell=True)
  
  if (SolverModule==True):
    
    plat = (platform.system().split('-')[0]).lower()
    if (plat=='windows' or plat=='darwin'):
     print ""
     print "<<ERROR>> SolverCuda module is supported only in Linux"
     print ""
    else:      
      print 'Building SolverCuda module...'
      print ''
      print ''
      raw_input('Recall that you need Cuda Toolkit\n' +
                'That version of Cuda requires that your gcc is less than 4.5 thus be sure \n'
                'that your gcc, g++ commands point to the proper version of gcc.\n'
                'Follow SolverCuda User Guide to assure that you use the correct gcc version.\n'
                'Press any key to continue...')
      print ''

      cudaTK = raw_input('Please insert the path to the CUDA TOOLKIT  (e.g. /usr/local/cuda)\n')
      cudaARC = raw_input('Please insert the CUDA architecture of your card (e.g. sm_20, sm_21). Default is sm_20\n')

      if (cudaARC==''):
        cudaARC = 'sm_20'

      if(os.path.isdir(cudaTK)==True):
       subprocess.call('cd '+solverDir+'build/ \n  cmake '+'-DCUDA_TOOLKIT_DIR='+cudaTK+' -DCUDA_ARCH='+cudaARC+' ..  \n make',shell=True)
      else:
       print "I will try to find the path to the CUDA TOOLKIT using CMAKE"
       subprocess.call('cd '+solverDir+'build/ \n  cmake -DCUDA_ARCH='+cudaARC+' .. \n make',shell=True)

      lib = 'libDelphiGpuSolver.so'
      solverLib = solverDir+'/lib/'+lib
  
      if (os.path.isfile(solverLib)==True):
       print '\n \n Compilation was successful!'
       print 'To install, get root priviliges, copy SolverModule/lib/libDelphiGpuSolver.so into your lib directory such as /usr/lib64, or if you do not have root privileges'
       print 'update LD_LIBRARY_PATH to point to libDelphiGpuSolver.so'       
       print ''
       raw_input('Press any key to continue...')
       print ''
       
       #print 'Installing SolverCuda library in /usr/lib or /usr/lib64'
       #install_cuda = 'cd ./lib \n chmod 777 ./%s \n rm -fr /usr/lib64/%s \n rm -fr /usr/lib/%s \n cp %s /usr/lib64 \n cp %s /usr/lib \n '%(lib,lib,lib,lib,lib)
       #subprocess.call(install_cuda,shell=True)

      else:
       print '<<ERROR>> Cannot find ./lib/libDelphiGpuSolver.so, compilation failed'
       print ''
       raw_input('Press any key to continue...')
       print ''
  
  if (NSmodule==True):
    print 'Building/Installing NanoShaper module...'
    subprocess.call('cd NanoShaper \n python setup.py',shell=True)

  print '\n'
  
if __name__ == "__main__":
  main()
